using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Data;
using System.Windows;
using System.Windows.Media.Imaging;
using Softix.Drawing.Diagram.Wpf.Core;
using System.Windows.Controls.Primitives;
using System.Windows.Shapes;
using Softix.Drawing.Diagram.Wpf.Resources;
using Softix.Orchestration.Core.Components;
using System.ComponentModel;
using System.Windows.Input;
using System.Collections;

namespace Softix.Drawing.Diagram.Wpf.Controls
{
	public class ComponentHost : ContentControl
	{
		internal const string OUTBOUND_PATHS_CONTROL = "OutboundPaths";
		internal const string COMPONENT_HOST_NAME_MASK = "host_{0}";

		private ModelEditor editor;
		private IAmComponent component;

		public ComponentHost(ModelEditor editor, IAmComponent component)
		{
			this.Style = StyleRepository.ComponentHostStyle;
			this.editor = editor;

			this.component = component;
			this.Name = String.Format(COMPONENT_HOST_NAME_MASK, this.component.ID.ToString().Replace("-", ""));
			this.component.PropertyChanged += this.ComponentPropertyChanged;
			this.component.ComponentChanged += this.ComponentChanged;
			this.ComponentChanged(this.component);
			this.SetComponentImage();
		}

		protected override void OnInitialized(EventArgs e)
		{
			base.OnInitialized(e);

			#region set component bindings

			Binding bLeft = new Binding();
			bLeft.Source = this.component;
			bLeft.Path = new PropertyPath("Left");
			bLeft.Mode = BindingMode.TwoWay;
			this.SetBinding(Canvas.LeftProperty, bLeft);

			Binding bTop = new Binding();
			bTop.Source = this.component;
			bTop.Path = new PropertyPath("Top");
			bTop.Mode = BindingMode.TwoWay;
			this.SetBinding(Canvas.TopProperty, bTop);

			Binding bWidth = new Binding();
			bWidth.Source = this.component;
			bWidth.Path = new PropertyPath("Width");
			bWidth.Mode = BindingMode.TwoWay;
			this.SetBinding(ContentControl.WidthProperty, bWidth);

			Binding bHeight = new Binding();
			bHeight.Source = this.component;
			bHeight.Path = new PropertyPath("Height");
			bHeight.Mode = BindingMode.TwoWay;
			this.SetBinding(ContentControl.HeightProperty, bHeight);

			#endregion
		}

		#region Component Changed Events

		private void ComponentChanged(IAmComponent component)
		{
			if (component != this.component)
				return;

			this.ComponentId = this.component.ID;

			// component's name
			this.ComponentName = this.component.Name;

			// component's path
			IAmComponentOutboundPath[] paths = this.component.OutboundPaths;
			if ((paths != null) && (paths.Length > 0))
			{
				List<ComponentPathInfo> pathInfos = new List<ComponentPathInfo>();
				for (int index = 0; index < paths.Length; index++)
					pathInfos.Add(new ComponentPathInfo(paths[index].Name, paths[index].ID));
				this.ComponentPaths = pathInfos.ToArray();
			}

			// set min height
			this.MinHeight = this.MinWidth + 10 * ((paths == null) ? 0 : paths.Length);
			if (this.component.Width == 0)
				this.component.Width = this.MinWidth;
			if (this.component.Height == 0)
				this.component.Height = this.MinHeight;
		}

		private void SetComponentImage()
		{
			BitmapImage bmp = new BitmapImage();
			bmp.BeginInit();
			bmp.CacheOption = BitmapCacheOption.OnLoad;
			System.Reflection.PropertyInfo pi = this.component.GetType().GetProperty("ImageResourcePath");
			bmp.StreamSource = pi.DeclaringType.Assembly.GetManifestResourceStream(this.component.ImageResourcePath);
			bmp.EndInit();
			this.ComponentImage = bmp;
		}

		private void ComponentPropertyChanged(object sender, PropertyChangedEventArgs e)
		{
			if (String.Compare(e.PropertyName, "name", true) == 0)
			{
				this.ComponentName = this.component.Name;
			}
			else if (String.Compare(e.PropertyName, "ImageResourcePath", true) == 0)
			{
				this.SetComponentImage();
			}
		}

		#endregion

		#region Properties

		public static readonly DependencyProperty ComponentIdProperty = DependencyProperty.Register(
			"ComponentId", typeof(Guid), typeof(ComponentHost));
		public static readonly DependencyProperty ComponentNameProperty = DependencyProperty.Register(
			"ComponentName", typeof(String), typeof(ComponentHost));
		public static readonly DependencyProperty ComponentImageProperty = DependencyProperty.Register(
			"ComponentImage", typeof(ImageSource), typeof(ComponentHost));
		public static readonly DependencyProperty ComponentPathsProperty = DependencyProperty.Register(
			"ComponentPaths", typeof(ComponentPathInfo[]), typeof(ComponentHost));

		public Guid ComponentId
		{
			get { return (Guid)this.GetValue(ComponentHost.ComponentIdProperty); }
			set { this.SetValue(ComponentHost.ComponentIdProperty, value); }
		}

		public string ComponentName
		{
			get { return (String)this.GetValue(ComponentHost.ComponentNameProperty); }
			set { this.SetValue(ComponentHost.ComponentNameProperty, value); }
		}

		public ImageSource ComponentImage
		{
			get { return (ImageSource)this.GetValue(ComponentHost.ComponentImageProperty); }
			set { this.SetValue(ComponentHost.ComponentImageProperty, value); }
		}

		public ComponentPathInfo[] ComponentPaths
		{
			get { return (ComponentPathInfo[])this.GetValue(ComponentHost.ComponentPathsProperty); }
			set { this.SetValue(ComponentHost.ComponentPathsProperty, value); }
		}

		internal ModelEditor Editor { get { return this.editor; } }
		internal IAmComponent Component { get { return this.component; } }

		#endregion

		#region Events

		internal static readonly RoutedEvent ShowContextMenuEvent =
			EventManager.RegisterRoutedEvent("ShowContextMenu", RoutingStrategy.Bubble, typeof(RoutedEventHandler), typeof(ComponentHost));

		internal event EventHandler ShowContextMenu
		{
			add { this.AddHandler(ComponentHost.ShowContextMenuEvent, value); }
			remove { this.RemoveHandler(ComponentHost.ShowContextMenuEvent, value); }
		}

		#endregion
	}
}
