using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Shapes;
using Softix.Orchestration.Core.Components;
using System.Windows.Media;
using System.Windows;
using System.Windows.Data;
using Softix.Drawing.Diagram.Wpf.Core;
using System.Windows.Controls.Primitives;
using System.Windows.Controls;
using Softix.Drawing.Diagram.Wpf.Resources;
using System.ComponentModel;

namespace Softix.Drawing.Diagram.Wpf.Controls
{
	public sealed class ComponentLinkHost : ContentControl
	{
		private ModelEditor editor;
		private IAmComponentLink link;

		private Line line = null;
		private ArrowDecorator decorator;

		public ComponentLinkHost(ModelEditor editor, IAmComponentLink link)
		{
			this.editor = editor;
			this.link = link;
			this.link.PropertyChanged += OnLinkPropertyChanged;

			// create decorator (arrow head)
			this.decorator = new ArrowDecorator();
			this.decorator.ArrowSize = 10;
			this.decorator.ArrowAngle = 25;
			this.AddChild(this.decorator);

			// create line
			this.line = new Line();
			this.line.Stroke = Brushes.Green;
			this.line.StrokeThickness = 1.0;
			this.decorator.Child = this.line;

			// set bindings and update line coordinates
			this.SetSourceBinding();
			this.SetTargetBinding();
			this.UpdateSourcePoint();
			this.UpdateTargetPoint();
		}

		private void OnLinkPropertyChanged(object sender, PropertyChangedEventArgs e)
		{
			if ((e.PropertyName == "SourceComponentPath") || (e.PropertyName == "SourceSide") ||
				(e.PropertyName == "SourceComponent"))
			{
				this.SetSourceBinding();
			}
			else if ((e.PropertyName == "DestinationSide") || (e.PropertyName == "TargetComponent"))
			{
				this.SetTargetBinding();
			}
		}

		#region Methods - Private Set Bindings

		private void SetSourceBinding()
		{
			BindingOperations.ClearBinding(this, ComponentLinkHost.SourcePointProperty);

			ComponentHost source = editor.FindComponentHostById(this.link.SourceComponent.ID);
			MultiBinding sourceBinding = new MultiBinding();
			sourceBinding.Bindings.Add(new Binding("Left"));
			sourceBinding.Bindings.Add(new Binding("Top"));
			sourceBinding.Bindings.Add(new Binding("Width"));
			sourceBinding.Bindings.Add(new Binding("Height"));
			foreach (Binding b in sourceBinding.Bindings)
				b.Source = this.link.SourceComponent;
			sourceBinding.Converter = new ComponentLinkSourceConverter();
			sourceBinding.ConverterParameter = new ComponentLinkSourceInfo(
				source,
				Array.Find<ComponentPathInfo>(source.ComponentPaths,
				delegate(ComponentPathInfo info)
				{
					bool tst = info.PathId == this.link.SourceComponentPath.ID;
					if (tst == true)
						info.IsLinked = true;
					return tst;
				}),
				this.link.SourceSide);
			BindingOperations.SetBinding(this, ComponentLinkHost.SourcePointProperty, sourceBinding);
		}

		private void SetTargetBinding()
		{
			BindingOperations.ClearBinding(this, ComponentLinkHost.DestinationPointProperty);

			ComponentHost target = editor.FindComponentHostById(this.link.TargetComponent.ID);
			MultiBinding targetBinding = new MultiBinding();
			targetBinding.Bindings.Add(new Binding("Left"));
			targetBinding.Bindings.Add(new Binding("Top"));
			targetBinding.Bindings.Add(new Binding("Width"));
			targetBinding.Bindings.Add(new Binding("Height"));
			foreach (Binding b in targetBinding.Bindings)
				b.Source = this.link.TargetComponent;
			targetBinding.Converter = new ComponentLinkDestinationConverter();
			targetBinding.ConverterParameter = new ComponentLinkDestinationInfo(
				target,
				this.link.DestinationSide);
			BindingOperations.SetBinding(this, ComponentLinkHost.DestinationPointProperty, targetBinding);
		}

		#endregion
		#region Methods - Overriden Property Changed & Helpers

		protected override void OnPropertyChanged(DependencyPropertyChangedEventArgs e)
		{
			base.OnPropertyChanged(e);

			if (this.line == null)
				return;

			if (e.Property == ComponentLinkHost.SourcePointProperty)
			{
				this.UpdateSourcePoint();
			}
			else if (e.Property == ComponentLinkHost.DestinationPointProperty)
			{
				this.UpdateTargetPoint();
			}
		}

		private void UpdateSourcePoint()
		{
			this.line.X1 = this.SourcePoint.X;
			this.line.Y1 = this.SourcePoint.Y;
		}

		private void UpdateTargetPoint()
		{
			this.line.X2 = this.DestinationPoint.X;
			this.line.Y2 = this.DestinationPoint.Y;
		}

		#endregion

		#region Properties

		public static readonly DependencyProperty SourcePointProperty = DependencyProperty.Register(
			"SourcePoint", typeof(Point), typeof(ComponentLinkHost));
		public static readonly DependencyProperty DestinationPointProperty = DependencyProperty.Register(
			"DestinationPoint", typeof(Point), typeof(ComponentLinkHost));

		public Point SourcePoint
		{
			get { return (Point)this.GetValue(ComponentLinkHost.SourcePointProperty); }
			set { this.SetValue(ComponentLinkHost.SourcePointProperty, value); }
		}

		public Point DestinationPoint
		{
			get { return (Point)this.GetValue(ComponentLinkHost.DestinationPointProperty); }
			set { this.SetValue(ComponentLinkHost.DestinationPointProperty, value); }
		}

		internal IAmComponentLink ComponentLink { get { return this.link; } }

		#endregion
	}
}
