using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Softix.Orchestration.Core.Models;
using Softix.Orchestration.Core.Components;

using Asm = System.Reflection.Assembly;
using System.ServiceModel;
using Softix.Orchestration.Core.Execution;
using System.CodeDom.Compiler;
using System.IO;
using Softix.Orchestration.Deployment.Internals;
using System.CodeDom;
using Softix.Orchestration.Core.Data;
using System.IO.Compression;
using System.Runtime.Serialization.Formatters.Binary;
using System.Runtime.Serialization;

namespace Softix.Orchestration.Deployment.WcfService
{
	public partial class WcfServiceDeployerUI : Form
	{
		private IAmModel model;
		private CodeNamespace code;

        public WcfServiceDeployerUI(IAmModel model)
        {
            InitializeComponent();

			this.model = model;
			IAmStartComponent[] components = this.model.StartComponents;
			if ((components != null) && (components.Length > 0))
			{
				foreach (IAmStartComponent start in components)
				{
					ListViewItem item = new ListViewItem();
					item.Text = start.Name;
					item.SubItems.Add("Execute" + Variable.GetPropertyName(start.Name));
					item.Tag = start;
					item.Checked = true;
					this.lvwStartComponents.Items.Add(item);
				}
			}

			this.ValidateChildren();
		}

		private Asm[] GetAssemblies()
		{
			List<Asm> list = new List<System.Reflection.Assembly>();
			list.Add(typeof(IAmModel).Assembly);
			list.Add(typeof(ModelExecutor).Assembly);
			foreach (IAmComponent component in this.model.Components)
			{
				if (component == null)
					continue;
				Asm asm = component.GetType().Assembly;
				if (list.Contains(asm) == false)
					list.Add(asm);
			}
			return list.ToArray();
		}

		private void btnGenerate_Click(object sender, EventArgs e)
		{
			// DirectoryInfo dir = new DirectoryInfo(Path.Combine(Path.GetTempPath(), Guid.NewGuid().ToString()));
			DirectoryInfo dir = new DirectoryInfo(this.txtAssemblyPath.Text + "Temp");
			dir.Create();

			try
			{
				// file: model
				FileInfo file = new FileInfo(Path.Combine(dir.FullName, "model"));
				using (Stream stream = file.Create())
				{
					using (GZipStream zip = new GZipStream(stream, CompressionMode.Compress))
					{
						(new BinaryFormatter()).Serialize(zip, this.model);
					}
				}

				// file: executor.cs
				FileInfo fileExecutor = new FileInfo(Path.Combine(dir.FullName, "executor.cs"));
				using (Stream stream = fileExecutor.Create())
				{
					List<StartComponentInfo> list = new List<StartComponentInfo>();
					foreach (ListViewItem item in this.lvwStartComponents.Items)
					{
						if ((item.Checked == true) && (item.Tag is IAmStartComponent))
						{
							list.Add(new StartComponentInfo((IAmStartComponent)item.Tag));
						}
					}
					if (list.Count == 0)
						throw new InvalidOperationException("No start component found.");

					StreamWriter writer = new StreamWriter(stream);
					writer.Write(GeneratorUtil.GenerateExecutorClass(this.model, true, this.txtClassNamespace.Text, this.txtClassName.Text, list.ToArray()));
					writer.Flush();
				}

				// file: modelreturn.cs
				FileInfo fileModelReturn = new FileInfo(Path.Combine(dir.FullName, "modelreturn.cs"));
				using (Stream stream = fileModelReturn.Create())
				{
					StreamWriter writer = new StreamWriter(stream);
					writer.Write(GeneratorUtil.GenerateModelReturnClass(this.model, true, this.txtClassNamespace.Text));
					writer.Flush();
				}

				FileInfo target = new FileInfo(this.txtAssemblyPath.Text);
				DirectoryInfo targetDir = target.Directory;
				if (targetDir.Exists == false)
					targetDir.Create();

				CodeDomProvider provider = Microsoft.CSharp.CSharpCodeProvider.CreateProvider("csharp");
				CompilerParameters options = new CompilerParameters();
				options.GenerateExecutable = true;
				options.OutputAssembly = this.txtAssemblyPath.Text;
				options.ReferencedAssemblies.Add(typeof(String).Assembly.Location);
				options.ReferencedAssemblies.Add(typeof(INotifyPropertyChanged).Assembly.Location);
				options.ReferencedAssemblies.Add(typeof(ServiceContractAttribute).Assembly.Location);
				options.ReferencedAssemblies.Add(typeof(DataContractAttribute).Assembly.Location);
				foreach (Asm asm in this.GetAssemblies())
				{
					options.ReferencedAssemblies.Add(asm.Location);
					if (this.chkIncludeAssemblies.Checked == true)
					{
						FileInfo targetAsmFile = new FileInfo(
							Path.Combine(targetDir.FullName, Path.GetFileName(asm.Location)));
						if (targetAsmFile.Exists == true)
							targetAsmFile.Delete();
						File.Copy(asm.Location, targetAsmFile.FullName);
					}
				}
				options.EmbeddedResources.Add(file.FullName);
				CompilerResults results = provider.CompileAssemblyFromFile(
					options,
					fileExecutor.FullName,
					fileModelReturn.FullName);

				if (results.Errors.HasErrors == true)
				{
					using (DummyViewer viewer = new DummyViewer())
					{
						viewer.ShowErrorCollection(results.Errors);
						viewer.ShowDialog(this);
					}
				}
				else
				{
					ConfigFileGenerator.GenerateConfigFile(
						this.txtAssemblyPath.Text,
						(String)this.cmbServiceProtocol.SelectedItem,
						this.txtClassNamespace.Text,
						this.txtClassName.Text,
						this.txtServicePort.Text);

					MessageBox.Show(this, "Assembly created!", "Deploy", MessageBoxButtons.OK, MessageBoxIcon.Information);
					this.Close();
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(this, ex.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
			finally { dir.Delete(true); }
		}

		private void txtClassName_Validating(object sender, CancelEventArgs e)
		{
			bool valid = ValidationUtil.ValidateClassName(this.txtClassName.Text);
			if (valid == true)
			{
				this.errorProvider.SetError(this.txtClassName, null);
				this.txtClassName.BackColor = SystemColors.Window;
			}
			else
			{
				this.errorProvider.SetError(this.txtClassName, "You must enter a valid class name!");
				this.txtClassName.BackColor = Color.Pink;
			}
		}

		private void txtClassNamespace_Validating(object sender, CancelEventArgs e)
		{
			bool valid = ValidationUtil.ValidateNamespace(this.txtClassNamespace.Text);
			if (valid == true)
			{
				this.errorProvider.SetError(this.txtClassNamespace, null);
				this.txtClassNamespace.BackColor = SystemColors.Window;
			}
			else
			{
				this.errorProvider.SetError(this.txtClassNamespace, "You must enter a valid namespace name!");
				this.txtClassNamespace.BackColor = Color.Pink;
			}
		}

		private void chkIncludeAssemblies_CheckedChanged(object sender, EventArgs e)
		{
			this.btnViewAssemblies.Enabled = this.chkIncludeAssemblies.Checked;
		}

		private void btnBrowse_Click(object sender, EventArgs e)
		{
			this.dlgBrowse.FileName = this.txtAssemblyPath.Text;
			if (this.dlgBrowse.ShowDialog(this) == DialogResult.OK)
			{
				this.txtAssemblyPath.Text = this.dlgBrowse.FileName;
			}
		}

		private void btnViewAssemblies_Click(object sender, EventArgs e)
		{
			using (DummyViewer viewer = new DummyViewer())
			{
				viewer.ShowAssemblyList(this.GetAssemblies());
				viewer.ShowDialog(this);
			}
		}

		private void txtClassName_TextChanged(object sender, EventArgs e)
		{
			this.txtInterfaceName.Text = String.Format("I{0}Service", this.txtClassName.Text);
			this.txtServiceUrl.Text = String.Format("{0}://localhost:{2}/{1}", this.cmbServiceProtocol.SelectedItem, this.txtClassName.Text, this.txtServicePort.Text);
		}

		private void WcfServiceDeployerUI_Load(object sender, EventArgs e)
		{
			this.cmbServiceProtocol.SelectedIndex = 0;
		}
    }
}