using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using Softix.Orchestration.Core.Components;
using Softix.Orchestration.Editor.Execution;
using Softix.Drawing.Diagram.Wpf;
using Softix.Orchestration.Core.Data;

namespace Softix.Orchestration.Editor.Controls
{
	public partial class ModelDebugger : UserControl
	{
        private static readonly Color EXECUTED_COLOR = Color.Gold;
        private static readonly Color LAST_EXECUTED_COLOR = Color.SteelBlue;
        private static readonly Color FAILED_COLOR = Color.OrangeRed;

        private ModelEditor editor;
        private DebugExecutor executor;
        private IAmComponent lastExecutedComponent;
        private IAmComponentLink lastExecutedComponentLink;

		public ModelDebugger(ModelEditor editor, IAmStartComponent component)
		{
			InitializeComponent();

            this.editor = editor;
            this.editor.DesignMode = false;

            this.executor = new DebugExecutor();
            this.executor.ComponentExecuted += this.OnComponentExecuted;
            this.executor.ComponentLinkExecuted += this.OnComponentLinkExecuted;
            this.executor.ComponentExecutionFailed += this.OnComponentExecutionFailed;
            this.executor.ModelExecutionEnded += this.OnModelExecutionEnded;
            this.executor.Run(component);
        }

        #region Private Event Handlers

        private void OnComponentLinkExecuted(object sender, DebugExecutor.ComponentLinkExecutionEventArgs e)
        {
            this.lvwComponents_SelectedIndexChanged(this.lvwComponents, EventArgs.Empty);

            if (this.lastExecutedComponentLink != null)
                this.editor.HighlightComponentLink(this.lastExecutedComponentLink, EXECUTED_COLOR);
            this.lastExecutedComponentLink = e.Link;
            this.editor.HighlightComponentLink(e.Link, LAST_EXECUTED_COLOR);
        }

        private void OnComponentExecuted(object sender, DebugExecutor.ComponentExecutionEventArgs e)
        {
            ListViewItem item = new ListViewItem(new string[] { e.Component.Name, e.ExecutionTime.ToString() });
            item.Tag = e.Component.ID;
            this.lvwComponents.Items.Add(item);
            item.Selected = true;

            if (this.lastExecutedComponent != null)
                this.editor.HighlightComponent(this.lastExecutedComponent, EXECUTED_COLOR);
            this.lastExecutedComponent = e.Component;
            this.editor.HighlightComponent(e.Component, LAST_EXECUTED_COLOR);
        }

        private void OnComponentExecutionFailed(object sender, DebugExecutor.ComponentExecutionFailedEventArgs e)
        {
            this.editor.HighlightComponent(e.Component, FAILED_COLOR);
            MessageBox.Show(this, e.Exception.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);

            if (this.lastExecutedComponent != null)
                this.editor.HighlightComponent(this.lastExecutedComponent, EXECUTED_COLOR);
        }

        private void OnModelExecutionEnded(object sender, EventArgs e)
        {
            this.btnNext.Enabled = false;
            this.btnStop.Enabled = false;
            this.btnRunToEnd.Enabled = false;

            this.btnEnd.Enabled = true;

            if (this.lastExecutedComponentLink != null)
                this.editor.HighlightComponentLink(this.lastExecutedComponentLink, EXECUTED_COLOR);
        }

        #endregion

        internal event EventHandler CloseDebugger;

        private void btnEnd_Click(object sender, EventArgs e)
        {
            this.editor.DesignMode = true;
            if (this.CloseDebugger != null)
                this.CloseDebugger(this, EventArgs.Empty);
        }

        private void btnNext_Click(object sender, EventArgs e)
        {
            this.executor.AutoPause = true;
            this.executor.ContinueExecution();
        }

        private void btnRunToEnd_Click(object sender, EventArgs e)
        {
            this.executor.AutoPause = false;
            this.executor.ContinueExecution();
        }

        private void btnStop_Click(object sender, EventArgs e)
        {
            this.executor.EndExecution();
        }

        private void lvwComponents_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (this.lvwComponents.SelectedItems.Count == 0)
                return;

            ListViewItem item = this.lvwComponents.SelectedItems[0];
            if (false == (item.Tag is Guid))
                return;

            VariableValue[] values = this.executor.CurrentDataStorage.GetOutputForComponent((Guid)item.Tag);
            VariablesPresenter presenter = new VariablesPresenter(values);
            this.pGrid.SelectedObject = presenter;
        }
    }
}
