using System;
using System.Collections.Generic;
using System.Text;
using Softix.Orchestration.Core.Execution;
using Softix.Orchestration.Core.Components;
using Softix.Orchestration.Core.Data;
using Softix.Orchestration.Core.Models;

namespace Softix.Orchestration.Editor.Execution
{
	public sealed class DebugExecutor
	{
        private bool autoPause = true;
        private DataStorage currentStorage = null;
        private IAmComponent currentComponent = null;

		public DebugExecutor()
		{
		}

		public void Run(IAmStartComponent component)
		{
			VariableValue[] initialData = ModelInputEditor.GetVariableValues(component);
			DataStorage storage = new DataStorage();
			if ((initialData != null) && (initialData.Length > 0))
			{
				foreach (VariableValue vv in initialData)
					storage.AddVariable(vv.ComponentId, vv.Name, vv.Value);
			}

            this.currentComponent = component;
            this.currentStorage = storage;

            try
            {
                DateTime dt = DateTime.Now;
                ComponentOutput output = component.StartExecution(initialData);
                this.OnComponentExecuted(component, DateTime.Now - dt);
                this.currentComponent = this.ProcessComponentOutput(storage, component, output);
                if (this.autoPause == false)
                    this.ContinueExecution();
            }
            catch (Exception ex)
            {
                this.ProcessComponentExecutionFail(component, ex);
            }
        }

        #region Methods

        public void ContinueExecution()
        {
            try
            {
                do
                {
                    DateTime dt = DateTime.Now;
                    InputVariableAttribute.SetInputVariables(this.currentComponent, this.currentStorage);
                    ComponentOutput output = this.currentComponent.Run(this.currentStorage);
                    this.OnComponentExecuted(this.currentComponent, DateTime.Now - dt);
                    IAmComponent nextComponent = this.ProcessComponentOutput(
                        this.currentStorage,
                        this.currentComponent,
                        output);

                    if (nextComponent == null)
                    {
                        // model execution ended
                        this.EndExecution();
                        break;
                    }
                    else
                    {
                        this.currentComponent = nextComponent;
                    }
                }
                while (this.autoPause == false);
            }
            catch (Exception ex)
            {
                this.ProcessComponentExecutionFail(this.currentComponent, ex);
            }
        }

        public void EndExecution()
        {
            this.currentComponent = null;
            this.OnModelExecutionEnded();
        }

        private IAmComponent ProcessComponentOutput(DataStorage storage, IAmComponent component, ComponentOutput output)
        {
            if (component is IAmEndComponent)
            {
                // execution ended
				VariableValue[] values = ((IAmEndComponent)component).EndExecution(storage);
				if ((values != null) && (values.Length > 0))
				{
					foreach (VariableValue value in values)
					{
						storage.AddVariable(value.ComponentId, value.Name, value.Value);
					}
				}
                return null;
            }

            // get path
            IAmComponentOutboundPath path = Array.Find<IAmComponentOutboundPath>(
                component.OutboundPaths,
                delegate(IAmComponentOutboundPath tmpPath)
                {
                    return (tmpPath.ID == output.OutboundPathID);
                });

            IAmModel model = component.ParentModel;

            if (path == null)
                throw new InvalidOperationException("Output path not found.");

            #region add variable values

            Variable[] vars = OutputVariableAttribute.GetOutputDataDefinitions(component, path);
            if ((vars != null) && (vars.Length > 0))
            {
                List<VariableValue> list = new List<VariableValue>();

                // add attribute-based output vars to cOut
                OutputVariableAttribute.AddToComponentOutput(component, path, output);
                // add variables
                foreach (Variable var in vars)
                {
                    object value = null;
                    try { value = output[var.Name]; }
                    catch { value = null; }

                    storage.AddVariable(component.ID, var.Name, value);
                }
            }

            #endregion

            // get next component
            IAmComponentLink[] links = model.GetOutboundLinksForComponent(component, path);
            if ((links == null) || (links.Length != 1))
                throw new InvalidOperationException("Incorrect number of outbound links for component '" + component.Name + "' found on path '" + path.Name + "'.");
            else
            {
                this.OnComponentLinkExecuted(links[0]);
                return links[0].TargetComponent;
            }
        }

        private void ProcessComponentExecutionFail(IAmComponent component, Exception ex)
        {
            this.OnComponentExecutionFailed(component, ex);
            this.OnModelExecutionEnded();
        }

        #endregion

        #region Events

        internal sealed class ComponentExecutionEventArgs : EventArgs
        {
            private IAmComponent component;
            private TimeSpan executionTime;

            internal ComponentExecutionEventArgs(IAmComponent component, TimeSpan executionTime)
            {
                this.component = component;
                this.executionTime = executionTime;
            }

            internal IAmComponent Component { get { return this.component; } }
            internal TimeSpan ExecutionTime { get { return this.executionTime; } }
        }

        internal sealed class ComponentLinkExecutionEventArgs : EventArgs
        {
            private IAmComponentLink link;

            internal ComponentLinkExecutionEventArgs(IAmComponentLink link)
            {
                this.link = link;
            }

            internal IAmComponentLink Link { get { return this.link; } }
        }

        internal sealed class ComponentExecutionFailedEventArgs : EventArgs
        {
            private IAmComponent component;
            private Exception exception;

            internal ComponentExecutionFailedEventArgs(IAmComponent component, Exception exception)
            {
                this.component = component;
                this.exception = exception;
            }

            internal IAmComponent Component { get { return this.component; } }
            internal Exception Exception { get { return this.exception; } }
        }

        internal event EventHandler<ComponentExecutionEventArgs> ComponentExecuted;
        internal event EventHandler<ComponentLinkExecutionEventArgs> ComponentLinkExecuted;
        internal event EventHandler<ComponentExecutionFailedEventArgs> ComponentExecutionFailed;
        internal event EventHandler ModelExecutionEnded;

        private void OnComponentExecuted(IAmComponent component, TimeSpan executionTime)
        {
            if (this.ComponentExecuted != null)
                this.ComponentExecuted(this, new ComponentExecutionEventArgs(component, executionTime));
        }

        private void OnComponentLinkExecuted(IAmComponentLink link)
        {
            if (this.ComponentLinkExecuted != null)
                this.ComponentLinkExecuted(this, new ComponentLinkExecutionEventArgs(link));
        }

        private void OnModelExecutionEnded()
        {
            if (this.ModelExecutionEnded != null)
                this.ModelExecutionEnded(this, EventArgs.Empty);
        }

        private void OnComponentExecutionFailed(IAmComponent component, Exception ex)
        {
            if (this.ComponentExecutionFailed != null)
                this.ComponentExecutionFailed(this, new ComponentExecutionFailedEventArgs(component, ex));
        }

        #endregion

        #region Properties

        public bool AutoPause
        {
            get { return this.autoPause; }
            set { this.autoPause = value; }
        }

        internal DataStorage CurrentDataStorage
        {
            get { return this.currentStorage; }
        }

        #endregion
    }
}
