using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using Softix.Orchestration.Core.Components;
using Softix.Orchestration.Core.Data;
using Softix.Orchestration.Core.Models;

namespace Softix.Orchestration.Core.Execution
{
    public sealed class ModelExecutor : IExecuteModels
    {
        private Thread executionThread;
        private IAmComponent currentComponent;
        private DataStorage currentStorage;

		private ModelExecutionStatus status;
		private VariableValue[] finalData = null;
		private Exception failReason = null;

        public ModelExecutor()
        {
			this.status = ModelExecutionStatus.NotStarted;
        }

        #region IExecuteModels Members

        public void StartExecution(IAmStartComponent component)
        {
            this.StartExecutionThread(component, null);
        }

        public void StartExecution(IAmStartComponent component, VariableValue[] initialData)
        {
            this.StartExecutionThread(component, initialData);
        }

        public ModelExecutionStatus QueryStatus()
        {
            return this.status;
        }

        #region events

        public event ModelExecutionStartedEventHandler ModelExecutionStarted;
        public event ModelExecutionFinishedEventHandler ModelExecutionFinished;
        public event ComponentExecutedEventHandler ComponentExecuted;
        public event ComponentExecutionFailedEventHandler ComponentExecutionFailed;

        private void OnModelExecutionStarted(IAmModel model, VariableValue[] initialData)
        {
			this.status = ModelExecutionStatus.Running;
            if (this.ModelExecutionStarted != null)
                this.ModelExecutionStarted(this, new ModelExecutionStartedEventArgs(model, initialData));
        }

        private void OnModelExecutionFinished(IAmModel model, VariableValue[] finalData)
        {
			this.status = ModelExecutionStatus.Ended;
			this.finalData = finalData;
            if (this.ModelExecutionFinished != null)
                this.ModelExecutionFinished(this, new ModelExecutionFinishedEventArgs(model, finalData));
        }

        private void OnModelExecutionFinished(IAmModel model, Exception exception)
        {
			this.status = ModelExecutionStatus.Failed;
			this.failReason = exception;
            if (this.ModelExecutionFinished != null)
                this.ModelExecutionFinished(this, new ModelExecutionFinishedEventArgs(model, exception));
        }

        private void OnComponentExecuted(IAmComponent component, IAmComponentLink outboundLink, TimeSpan executionTime)
        {
            if (this.ComponentExecuted != null)
                this.ComponentExecuted(this, new ComponentExecutedEventArgs(component, outboundLink, executionTime));
        }

        private void OnComponentExecutionFailed(IAmComponent component, Exception exception)
        {
            if (this.ComponentExecutionFailed != null)
                this.ComponentExecutionFailed(this, new ComponentExecutionFailedEventArgs(component, exception));
        }

        #endregion
        #endregion

        #region Methods - Privates

        private static void ValidateModelInitialData(VariableValue[] initialData, IAmModel model)
        {
            Variable[] inputVariables = model.InitialVariables;
            if ((inputVariables != null) && (inputVariables.Length > 0))
            {
                if (initialData == null)
                    throw new InvalidOperationException("You must specify the initial data for this model.");

                foreach (Variable variable in inputVariables)
                {
                    if (Array.Exists<VariableValue>(initialData,
                        delegate(VariableValue vv)
                        {
                            return (vv.Name == variable.Name);
                        }) == false)
                    {
                        throw new InvalidOperationException("Initial value of variable '" + variable.Name + "' not specified.");
                    }
                }
            }
        }

        #endregion
        #region Methods - Model Execution

        private Thread StartExecutionThread(IAmStartComponent component, VariableValue[] initialData)
        {
            if (this.QueryStatus() != ModelExecutionStatus.NotStarted)
                throw new InvalidOperationException("This executor is not in Idle state.");

            if (component.ParentModel == null)
                throw new InvalidOperationException("Component is not part of a model.");

            IAmModel model = component.ParentModel;
            ModelExecutor.ValidateModelInitialData(initialData, model);

            try
            {
                DateTime dt = DateTime.Now;
                ComponentOutput output = component.StartExecution(initialData);
                this.currentStorage = new DataStorage();
                this.currentComponent = this.ProcessComponentOutput(this.currentStorage, component, output, dt);

                this.executionThread = new Thread(new ParameterizedThreadStart(this.ExecuteModelInThread));
                this.executionThread.Start(component.ID);
            }
            catch (Exception ex)
            {
                this.OnComponentExecutionFailed(component, ex);
            }
            return this.executionThread;
        }

        private void ExecuteModelInThread(object parameter)
        {
            if (false == (parameter is Guid))
            {
                Guid[] ids = this.currentStorage.GetComponentIDs();
                if ((ids != null) && (ids.Length > 0))
                    this.ExecuteModelInThread(ids[0]);
                else
                    this.ExecuteModelInThread(Guid.Empty);
            }
            else
            {
                this.ExecuteModelInThread((Guid)parameter);
            }
        }

        private void ExecuteModelInThread(Guid startComponentId)
        {
            IAmModel model = this.currentComponent.ParentModel;
            try
            {
                this.OnModelExecutionStarted(model, this.currentStorage.GetOutputForComponent(startComponentId));
                Guid lastComponentId;
                do
                {
                    DateTime dt = DateTime.Now;
                    InputVariableAttribute.SetInputVariables(this.currentComponent, this.currentStorage);
                    ComponentOutput output = this.currentComponent.Run(this.currentStorage);
                    IAmComponent nextComponent = this.ProcessComponentOutput(
                        this.currentStorage,
                        this.currentComponent,
                        output, dt);

                    lastComponentId = this.currentComponent.ID;
                    if ((nextComponent == null) && (false == (this.currentComponent is IAmEndComponent)))
                        throw new InvalidOperationException("Model ended with a component that is not a end component.");
                    this.currentComponent = nextComponent;
                }
                while (this.currentComponent != null);
                this.OnModelExecutionFinished(model, this.currentStorage.GetOutputForComponent(lastComponentId));
            }
            catch (Exception ex)
            {
                this.OnComponentExecutionFailed(this.currentComponent, ex);
                this.OnModelExecutionFinished(model, ex);
            }
        }

        private IAmComponent ProcessComponentOutput(DataStorage storage, IAmComponent component, ComponentOutput output, DateTime startTime)
        {
            if (component is IAmEndComponent)
            {
				// execution ended
				VariableValue[] values = ((IAmEndComponent)component).EndExecution(storage);
				if ((values != null) && (values.Length > 0))
				{
					foreach (VariableValue value in values)
					{
						storage.AddVariable(value.ComponentId, value.Name, value.Value);
					}
				}
				return null;
            }

            // get path
            IAmComponentOutboundPath path = Array.Find<IAmComponentOutboundPath>(
                component.OutboundPaths,
                delegate(IAmComponentOutboundPath tmpPath)
                {
                    return (tmpPath.ID == output.OutboundPathID);
                });

            IAmModel model = component.ParentModel;

            if (path == null)
                throw new InvalidOperationException("Output path not found.");

            #region add variable values

            Variable[] vars = OutputVariableAttribute.GetOutputDataDefinitions(component, path);
            if ((vars != null) && (vars.Length > 0))
            {
                List<VariableValue> list = new List<VariableValue>();

                // add attribute-based output vars to cOut
                OutputVariableAttribute.AddToComponentOutput(component, path, output);
                // add variables
                foreach (Variable var in vars)
                {
                    object value = null;
                    try { value = output[var.Name]; }
                    catch { value = null; }

                    storage.AddVariable(component.ID, var.Name, value);
                }
            }

            #endregion

            // get next component
            IAmComponentLink[] links = model.GetOutboundLinksForComponent(component, path);
            if ((links == null) || (links.Length != 1))
                throw new InvalidOperationException("Incorrect number of outbound links for component '" + component.Name + "' found on path '" + path.Name + "'.");
            else
            {
                this.OnComponentExecuted(component, links[0], DateTime.Now - startTime);
                return links[0].TargetComponent;
            }
        }

        #endregion
        #region Methods - Static

        public static VariableValue[] ExecuteModelSync(IAmStartComponent component, VariableValue[] initialData)
        {
			ModelExecutor executor = new ModelExecutor();
			Thread t = executor.StartExecutionThread(component, initialData);
			if ((t != null) && (t.IsAlive == true))
				t.Join();
			if ((executor.status == ModelExecutionStatus.Failed) && (executor.failReason != null))
				throw executor.failReason;
			else if (executor.status == ModelExecutionStatus.Ended)
				return executor.finalData;
			throw new InvalidOperationException("Model execution was not finished with success.");
        }

        #endregion
    }
}
