using System;
using System.Collections.Generic;
using System.Text;

using Softix.Orchestration.Core.Validation;
using Softix.Orchestration.Core.Models;
using System.Runtime.Serialization;
using System.ComponentModel;

namespace Softix.Orchestration.Core.Data
{
    [Serializable]
    public sealed class Variable : ValidationBase, ISupportValidation, ISerializable, INotifyPropertyChanged
    {
        public Variable()
        { }

        public Variable(Guid sourceComponentID, String name, IAmDataType dataType)
        {
            this.sourceComponentID = sourceComponentID;
            this.name = name;
            this.typeHandler = dataType;
        }

        #region public Guid SourceComponentID { get; set; }

        private Guid sourceComponentID;

        [ Browsable(false) ]
        public Guid SourceComponentID
        {
            get { return sourceComponentID; }
            set
            {
                if (this.sourceComponentID != value)
                {
                    this.sourceComponentID = value;
                    base.RunValidation();
                    this.OnPropertyChanged("SourceComponentId");
                }
            }
        }

        #endregion
        #region public String Name { get; set; }

        private String name;

        public String Name
        {
            get { return name; }
            set
            {
                if (this.name != value)
                {
                    name = value;
                    base.RunValidation();
                    this.OnPropertyChanged("Name");
                }
            }
        }

        #endregion

        #region public Type Type { get; set; }

        private IAmDataType typeHandler = null;

        [ TypeConverter(typeof(VariableTypeConverter)) ]
        public IAmDataType DataType
        {
            get { return this.typeHandler; }
            set
            {
                if (this.typeHandler != value)
                {
                    this.typeHandler = value;
                    base.RunValidation();
                    this.OnPropertyChanged("DataType");
                }
            }
        }

        #endregion
        #region public bool IsArray { get; }

        [ ReadOnly(true) ]
		[Browsable(false)]
		public bool IsArray
        {
            get { return (this.typeHandler is IAmArrayDataType); }
        }

        #endregion

        #region public Boolean IsLoopVariable { get; set; }

        private bool loopVariable = false;

		[ Browsable(false) ]
        public bool IsLoopVariable
        {
            get { return this.loopVariable; }
            set { this.loopVariable = value; }
        }

        #endregion

        #region ISOSupportsValidation Members

        private IEnumerable<ValidationBreak> BaseValidationBreaks
        {
            get { return base.GetValidationBreaks(); }
        }

        protected override IEnumerable<ValidationBreak> GetValidationBreaks()
        {
            // add base breaks
            foreach (ValidationBreak vb in this.BaseValidationBreaks)
                yield return vb;

            if (String.IsNullOrEmpty(this.name) == true)
                yield return new ValidationBreak(this,
                    "Property 'name' cannot be empty.",
                    "You must specify a name for this data definition.");

            if (this.typeHandler == null)
                yield return new ValidationBreak(this,
                    "Property 'type' cannot be null.",
                    "You must specify a type for this data definition.");

        }

        #endregion

        #region ISerializable Members

        public Variable(SerializationInfo info, StreamingContext context)
        {
            this.name = info.GetString("_.name");
            this.sourceComponentID = (Guid)info.GetValue("_.componentId", typeof(Guid));

			try
			{
				this.typeHandler = DataTypeFactory.GetDataType((Type)info.GetValue("_.dataType", typeof(Type)));
			}
			catch { }
        }

        public void GetObjectData(SerializationInfo info, StreamingContext context)
        {
            info.AddValue("_.name", this.Name);
            info.AddValue("_.componentId", this.SourceComponentID);
            info.AddValue("_.dataType", this.DataType.DataType);
        }

        #endregion

        #region INotifyPropertyChanged Members

        [ field: NonSerialized ]
        public event PropertyChangedEventHandler PropertyChanged;

        private void OnPropertyChanged(string propertyName)
        {
            if (this.PropertyChanged != null)
                this.PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }

        protected override void OnValidationChanged()
        {
            base.OnValidationChanged();
            this.OnPropertyChanged("IsValid");
        }

        #endregion

		#region Methods - Code Generators

        [ Browsable(false) ]
		public string FieldName
		{
			get { return GetFieldName(this.name); }
		}

        [Browsable(false)]
        public string PropertyName
		{
			get { return GetPropertyName(this.name); }
		}

		public static string GetFieldName(string name)
		{
			string camel = Variable.CamelCaseName(name);
			if (camel.Length == 0)
				return "";
			if (camel.Length == 1)
				return camel.ToLower();
			return Char.ToLower(camel[0]) + camel.Substring(1);
		}

		public static string GetPropertyName(string name)
		{
			string camel = Variable.CamelCaseName(name);
			if (camel.Length == 0)
				return "";
			if (camel.Length == 1)
				return camel.ToUpper();
			return Char.ToUpper(camel[0]) + camel.Substring(1);
		}

		public static string CamelCaseName(string name)
		{
			string[] parts = name.Split('\t', '\n', ' ');
			StringBuilder sb = new StringBuilder();
			foreach (string part in parts)
			{
				string tmp = part.Trim();
				if (tmp.Length == 0)
					continue;
				if (tmp.ToUpper() == tmp)
					sb.Append(tmp);
				else if (tmp.Length == 1)
					sb.Append(tmp.ToUpper());
				else
				{
					sb.Append(Char.ToUpper(tmp[0]));
					sb.Append(tmp.Substring(1).ToLower());
				}
			}
			return sb.ToString();
		}

		#endregion

    }
}
