using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.Serialization;

using Softix.Orchestration.Core.Components;
using Softix.Orchestration.Core.Components.Events;
using Softix.Orchestration.Core.Data;
using Softix.Orchestration.Core.Models.Events;
using Softix.Orchestration.Core.Validation;
using System.ComponentModel;

namespace Softix.Orchestration.Core.Models
{
    [ Serializable ]
    public abstract class ModelBase : ValidationBase, IAmModel, IDeserializationCallback
    {
        private List<IAmComponent> components = new List<IAmComponent>();
        private List<IAmComponentLink> links = new List<IAmComponentLink>();

        #region public String Name { get; set; }

        private String name;

        public String Name
        {
            get { return name; }
            set { name = value; }
        }

        #endregion

        #region IAmModel Members, components

        public void AddComponent(IAmComponent component)
        {
            IAmComponent tmpComp = this.GetComponent(component.ID);
            if (tmpComp == component)
                return;
            if (tmpComp != null)
                throw new InvalidOperationException("Duplicate component ID found.");

            // add component
            if (component.ParentModel != null)
                component.ParentModel.RemoveComponent(component);
            component.ParentModel = this;
            this.components.Add(component);
            this.OnComponentAdded(component);
        }

        public void RemoveComponent(Guid componentID)
        {
            IAmComponent tmpComp = this.GetComponent(componentID);
            if (tmpComp != null)
                this.RemoveComponent(tmpComp);
        }

        public void RemoveComponent(IAmComponent component)
        {
            if (component.ParentModel == this)
            {
                // delete inbound links
                IAmComponentLink[] inLinks = this.GetInboundLinksForComponent(component);
                if ((inLinks != null) && (inLinks.Length > 0))
                {
                    foreach (IAmComponentLink link in inLinks)
                        this.RemoveComponentLink(link);
                }
                // delete inbound links
                IAmComponentLink[] outLinks = this.GetOutboundLinksForComponent(component);
                if ((outLinks != null) && (outLinks.Length > 0))
                {
                    foreach (IAmComponentLink link in outLinks)
                        this.RemoveComponentLink(link);
                }
                // remove component
                component.ParentModel = null;
                this.components.Remove(component);
                this.OnComponentRemoved(component);
            }
        }

        public IAmComponent GetComponent(Guid componentID)
        {
            return this.components.Find(
                delegate(IAmComponent comp)
                {
                    return (comp.ID == componentID);
                });
        }

		[Browsable(false)]
		public IAmComponent[] Components
        {
            get { return this.components.ToArray(); }
        }

        #region events

        [field: NonSerialized]
        public event ModelComponentHandler ComponentAdded;
        [field: NonSerialized]
        public event ModelComponentHandler ComponentRemoved;
        [field: NonSerialized]
        public event ModelComponentHandler ComponentChanged;

        protected virtual void OnComponentAdded(IAmComponent component)
        {
            this.RunValidation();
            if (this.ComponentAdded != null)
                this.ComponentAdded(this, component);
        }

        protected virtual void OnComponentRemoved(IAmComponent component)
        {
            this.RunValidation();
            if (this.ComponentRemoved != null)
                this.ComponentRemoved(this, component);
        }

        protected virtual void OnComponentChanged(IAmComponent component)
        {
            this.RunValidation();
            if (this.ComponentChanged != null)
                this.ComponentChanged(this, component);
        }

        #endregion

        #endregion

        #region IAmModel Members, component links

        public void AddComponentLink(IAmComponentLink link)
        {
            IAmComponentLink tmpLink = this.GetComponentLink(link.ID);
            if (tmpLink == link)
                return;
            if (tmpLink != null)
                throw new InvalidOperationException("Duplicate component link found.");

            // test if source and target are in this model
            IAmComponent source = this.GetComponent(link.SourceComponent.ID);
            if (source == null)
                throw new InvalidOperationException("Source component not found in model.");
            if (Array.Find<IAmComponentOutboundPath>(source.OutboundPaths,
                delegate(IAmComponentOutboundPath tmpPath)
                {
                    return (tmpPath.ID == link.SourceComponentPath.ID);
                }) == null)
                throw new InvalidOperationException("Source component path not found.");
            if (this.GetComponent(link.TargetComponent.ID) == null)
                throw new InvalidOperationException("Target component not found.");

            // search for duplicate link
            tmpLink = this.links.Find(
                delegate(IAmComponentLink tmp)
                {
                    return (
                        (tmp.SourceComponent.ID == link.SourceComponent.ID) &&
                        (tmp.SourceComponentPath.ID == link.SourceComponentPath.ID) &&
                        (tmp.TargetComponent.ID == link.TargetComponent.ID));
                });
            if (tmpLink != null)
                throw new InvalidOperationException("Duplicate link found.");

            // add link
            if (link.ParentModel != null)
                link.ParentModel.RemoveComponentLink(link);
            link.ParentModel = this;
            this.links.Add(link);
            this.OnComponentLinkAdded(link);
            
            // validate source & target
            if (link.SourceComponent is ISupportValidation)
                ((ISupportValidation)link.SourceComponent).RunValidation();
            if (link.TargetComponent is ISupportValidation)
                ((ISupportValidation)link.TargetComponent).RunValidation();
        }

        public void RemoveComponentLink(Guid linkID)
        {
            IAmComponentLink tmpLink = this.GetComponentLink(linkID);
            if (tmpLink != null)
                this.RemoveComponentLink(tmpLink);
        }

        public void RemoveComponentLink(IAmComponentLink link)
        {
            if (link.ParentModel == this)
                link.ParentModel = null;
            this.links.Remove(link);
            this.OnComponentLinkRemoved(link);

            // validate source & target
            if (link.SourceComponent is ISupportValidation)
                ((ISupportValidation)link.SourceComponent).RunValidation();
            if (link.TargetComponent is ISupportValidation)
                ((ISupportValidation)link.TargetComponent).RunValidation();
        }

        public IAmComponentLink GetComponentLink(Guid linkID)
        {
            return this.links.Find(
                delegate(IAmComponentLink link)
                {
                    return (link.ID == linkID);
                });
        }

        #region component inbout/outbound links

        public IAmComponentLink[] GetOutboundLinksForComponent(Guid componentID)
        {
            return this.links.FindAll(
                delegate(IAmComponentLink link)
                {
                    return (link.SourceComponent.ID == componentID);
                }).ToArray();
        }

        public IAmComponentLink[] GetOutboundLinksForComponent(IAmComponent component)
        {
            return this.GetOutboundLinksForComponent(component.ID);
        }

        public IAmComponentLink[] GetOutboundLinksForComponent(Guid componentID, Guid pathID)
        {
            return this.links.FindAll(
                delegate(IAmComponentLink link)
                {
                    return ((link.SourceComponent.ID == componentID) &&
                        (link.SourceComponentPath.ID == pathID));
                }).ToArray();
        }

        public IAmComponentLink[] GetOutboundLinksForComponent(IAmComponent component, IAmComponentOutboundPath path)
        {
            return this.GetOutboundLinksForComponent(component.ID, path.ID);
        }

        public IAmComponentLink[] GetInboundLinksForComponent(Guid componentID)
        {
            return this.links.FindAll(
                delegate(IAmComponentLink link)
                {
                    return (link.TargetComponent.ID == componentID);
                }).ToArray();
        }

        public IAmComponentLink[] GetInboundLinksForComponent(IAmComponent component)
        {
            return this.GetInboundLinksForComponent(component.ID);
        }

        #endregion

        #region events

        [field: NonSerialized]
        public event ModelComponentLinkHandler ComponentLinkAdded;
        [field: NonSerialized]
        public event ModelComponentLinkHandler ComponentLinkRemoved;
        [field: NonSerialized]
        public event ModelComponentLinkHandler ComponentLinkChanged;

        protected virtual void OnComponentLinkAdded(IAmComponentLink link)
        {
            this.RunValidation();
            if (this.ComponentLinkAdded != null)
                this.ComponentLinkAdded(this, link);
        }

        protected virtual void OnComponentLinkRemoved(IAmComponentLink link)
        {
            this.RunValidation();
            if (this.ComponentLinkRemoved != null)
                this.ComponentLinkRemoved(this, link);
        }

        protected virtual void OnComponentLinkChanged(IAmComponentLink link)
        {
            this.RunValidation();
            if (this.ComponentLinkChanged != null)
                this.ComponentLinkChanged(this, link);
        }

        #endregion

		public bool CanLinkComponents(
			IAmComponent sourceComponent,
			IAmComponentOutboundPath sourceComponentPath,
			IAmComponent destinationComponent)
		{
			if (sourceComponent == destinationComponent)
				return false;

			IAmComponentLink[] links = this.GetOutboundLinksForComponent(sourceComponent.ID);
			if ((links != null) && (links.Length > 0))
			{
				foreach (IAmComponentLink link in links)
				{
					if (link.SourceComponentPath == sourceComponentPath)
						return false;
				}
			}

			return true;
		}

		[Browsable(false)]
		public IAmComponentLink[] ComponentLinks
        {
            get { return this.links.ToArray(); }
        }

        #endregion

        #region IAmModel Members, data for links

        public Variable[] GetVariablesForComponentLink(IAmComponentLink link)
        {
            return this.GetVariablesForComponentLink(link.ID, link.TargetComponent.ID);
        }

        protected void SearchVariables(Guid linkID, Guid targetComponentId, List<Guid> visited, VariableList variables)
        {
            IAmComponentLink link = this.GetComponentLink(linkID);
            if (link == null)
                throw new InvalidOperationException("Component link not found in model.");
            if (visited.Contains(link.ID) == true)
            {
                variables.RemoveAll(
                    delegate(Variable var)
                    {
                        return ((var.IsLoopVariable == true) && (var.SourceComponentID == link.TargetComponent.ID));
                    });
                return;
            }
            visited.Add(link.ID);

            Variable[] sourceCompDefs = OutputVariableAttribute.GetOutputDataDefinitions(
                link.SourceComponent,
                link.SourceComponentPath);
            if ((sourceCompDefs != null) && (sourceCompDefs.Length > 0))
            {
                variables.AddRange(sourceCompDefs);
            }

            foreach (IAmComponentLink tmpLink in this.GetInboundLinksForComponent(link.SourceComponent))
            {
                this.SearchVariables(tmpLink.ID, targetComponentId, visited, variables);
            }
        }

        public Variable[] GetVariablesForComponentLink(Guid linkID, Guid targetComponentId)
        {
            VariableList list = new VariableList();
            this.SearchVariables(linkID, targetComponentId, new List<Guid>(), list);
            return list.UniqueVariables.ToArray();
        }

        #endregion

        #region IAmModel Members, data for components

        public Variable[] GetVariablesForComponent(Guid componentID)
        {
            IAmComponent component = this.GetComponent(componentID);
            if (component == null)
                throw new InvalidOperationException("Component not found in model.");

            VariableList list = new VariableList();

            IAmComponentLink[] inbound = this.GetInboundLinksForComponent(component.ID);
            if (inbound != null)
            {
                foreach (IAmComponentLink link in inbound)
                {
                    Variable[] linkVars = this.GetVariablesForComponentLink(link);
                    if ((linkVars != null) && (linkVars.Length > 0))
                        list.AddRange(linkVars);
                }
            }

            return list.UniqueVariables.ToArray();
        }

        public Variable[] GetVariablesForComponent(IAmComponent component)
        {
            return this.GetVariablesForComponent(component.ID);
        }

        #endregion

        #region IAmModel Members, data

        #region public Variable[] InitialVariables { get; set; }

        private Variable[] initialVariables;

		[Browsable(false)]
		public Variable[] InitialVariables
        {
            get { return (initialVariables != null) ? initialVariables : new Variable[0]; }
            set
            {
                initialVariables = value;
                this.OnInitialVariablesChanged();
            }
        }

        #endregion
        #region public Variable[] FinalVariables { get; set; }

        private Variable[] finalVariables;

		[Browsable(false)]
		public Variable[] FinalVariables
        {
            get { return (finalVariables != null) ? finalVariables : new Variable[0]; }
            set
            {
                finalVariables = value;
                this.OnFinalVariablesChanged();
            }
        }

        #endregion
    
        [field: NonSerialized]
        public event ModelDataChanged InitialVariablesChanged;
        [field: NonSerialized]
        public event ModelDataChanged FinalVariablesChanged;

        protected void OnInitialVariablesChanged()
        {
            if (this.InitialVariablesChanged != null)
                this.InitialVariablesChanged(this);
        }

        protected void OnFinalVariablesChanged()
        {
            if (this.FinalVariablesChanged != null)
                this.FinalVariablesChanged(this);
        }

        #endregion

        #region validation

        private IEnumerable<ValidationBreak> BaseValidationBreaks
        {
            get { return base.GetValidationBreaks(); }
        }

        protected override IEnumerable<ValidationBreak> GetValidationBreaks()
        {
            // add base reasons
            foreach (ValidationBreak validationBreak in this.BaseValidationBreaks)
                yield return validationBreak;

            // validate initial data definitions
            if (this.initialVariables != null)
            {
                foreach (Variable var in this.initialVariables)
                {
                    // validate definitions
                    var.RunValidation();
                    if (var.IsValid == false)
                    {
                        foreach (ValidationBreak vb in var.ValidationBreaks)
                            yield return vb;
                    }
                }
            }

            // validate final data definitions
            if (this.finalVariables != null)
            {
                foreach (Variable def in this.finalVariables)
                {
                    // validate definitions
                    def.RunValidation();
                    if (def.IsValid == false)
                    {
                        foreach (ValidationBreak vb in def.ValidationBreaks)
                            yield return vb;
                    }
                }
            }

            // validate components
            foreach (IAmComponent component in this.components)
            {
                if (component is ISupportValidation)
                {
                    ISupportValidation valid = (ISupportValidation)component;
                    valid.RunValidation();
                    if (valid.IsValid == false)
                    {
                        foreach (ValidationBreak vb in valid.ValidationBreaks)
                            yield return vb;
                    }
                }
            }

            // validate links (??)
            foreach (IAmComponentLink link in this.links)
            {
                if (link is ISupportValidation)
                {
                    ISupportValidation valid = (ISupportValidation)link;
                    valid.RunValidation();
                    if (valid.IsValid == false)
                    {
                        foreach (ValidationBreak vb in valid.ValidationBreaks)
                            yield return vb;
                    }
                }
            }
        }

        #endregion

        #region IDeserializationCallback Members

        public virtual void OnDeserialization(object sender)
        {
            // gigi was here
            foreach(IAmComponent component in this.components)
                component.ComponentChanged += new ComponentEvent(OnComponentChanged);
            this.RunValidation();
        }

        #endregion

		[ Browsable(false) ]
        public IAmStartComponent[] StartComponents
        {
            get
            {
                return this.components.FindAll(
                    delegate(IAmComponent comp)
                    {
						return (comp is IAmStartComponent);
                    }).ConvertAll<IAmStartComponent>(
					delegate(IAmComponent comp)
					{
						return (IAmStartComponent)comp;
					}).ToArray();
            }
        }
    }
}
