using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;

using Softix.Drawing.Diagram.Core;
using Softix.Drawing.Diagram.Core.Events;
using Softix.Drawing.Diagram.Internals;
using DiagramObjectInfoListItem = System.Collections.Generic.KeyValuePair<
    Softix.Drawing.Diagram.Core.IDiagramObject,
    Softix.Drawing.Diagram.Internals.DiagramObjectInfo>;
using DiagramObjectInfoPortListItem = System.Collections.Generic.KeyValuePair<
    Softix.Drawing.Diagram.Core.IDiagramObjectPort,
    Softix.Drawing.Diagram.Internals.DiagramObjectPortInfo>;

namespace Softix.Drawing.Diagram
{
    public partial class DiagramView : UserControl
    {
        private DiagramContext context = null;

        public DiagramView()
        {
            InitializeComponent();
            this.Font = new Font("Tahoma", 8.25f, FontStyle.Regular, GraphicsUnit.Point);
            this.DoubleBuffered = true;
        }

        #region public IDiagram Diagram { get; set; }

        private IDiagram diagram;

        public IDiagram Diagram
        {
            get { return diagram; }
            set
            {
                EventHandler redrawDiagam = new EventHandler(this.RedrawDiagram);
                if (this.diagram != null)
                    this.diagram.ShouldRedraw -= redrawDiagam;

                this.diagram = value;
                if (this.diagram != null)
                {
                    this.context = new DiagramContext(this, diagram);
                    this.diagram.ShouldRedraw += redrawDiagam;
                }
                else
                    this.context = null;
            }
        }

        #endregion

        #region redraw diagram

        private void RedrawDiagram(object sender, EventArgs args)
        {
            if (sender != this.diagram)
                return; // TODO: maybe throw an error...:-?
            this.Invalidate();
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);
            if (this.context == null)
                return;

            e.Graphics.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
            DiagramObjectInfoList dic = new DiagramObjectInfoList();

            //////////
            // draw objects
            this.context.DrawObjects(e.Graphics);

            //////////
            // draw links
            this.context.DrawLinks(e.Graphics);

            //////////
            // draw special ui
            this.context.DrawSpecialUI(e.Graphics);

            // draw error message
            if (String.IsNullOrEmpty(this.diagram.ErrorMessage) == false)
            {
                string msg = this.diagram.ErrorMessage;
                Font msgFont = new Font("Verdana", 9.0f, FontStyle.Bold, GraphicsUnit.Pixel);
                StringFormat msgFormat = new StringFormat(StringFormatFlags.FitBlackBox);
                msgFormat.Alignment = StringAlignment.Near;
                msgFormat.LineAlignment = StringAlignment.Center;

                Size msgSize = Size.Ceiling(e.Graphics.MeasureString(msg, msgFont, this.Width - 10, msgFormat));
                Rectangle msgBg = new Rectangle(
                    0, this.Height - msgSize.Height - 6,
                    this.Width, msgSize.Height + 6);
                e.Graphics.FillRectangle(Brushes.LightYellow, msgBg);

                Rectangle msgBounds = new Rectangle(
                    5, this.Height - msgSize.Height - 3,
                    msgSize.Width, msgSize.Height);
                e.Graphics.DrawString(msg, msgFont, Brushes.OrangeRed, msgBounds, msgFormat);
            }
        }

        #endregion

        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);
            if (this.context == null)
                return;

            if ((this.context.DiagramState == DiagramState.None) ||
                (this.context.DiagramState == DiagramState.HoverEntity))
            {
                IHoverObjectInfo newHover = null;
                foreach (DiagramObjectInfo info in this.context.DiagramObjectInfos)
                {
                    newHover = info.TestHover(e.Location);
                    if (newHover != null)
                    {
                        this.context.SetHoveringEntity(newHover);
                        break;
                    }
                }
                if (newHover == null)
                    this.context.ClearState();
            }
            else if (this.context.DiagramState == DiagramState.MoveObject)
            {
                this.context.UpdateMovingObject(e.Location);
            }
            else if (this.context.DiagramState == DiagramState.DragPort)
            {
                DiagramObjectTarget newTarget = null;
                foreach (DiagramObjectInfo info in this.context.DiagramObjectInfos)
                {
                    if (info.DiagramObject == this.context.DragPort.DiagramObject)
                        continue;

                    IHoverObjectInfo newHover = info.TestHover(e.Location, false);
                    if (newHover == null)
                        continue;
                    newTarget = new DiagramObjectTarget(info);

                    if ((context.DragPortTarget == null) ||
                        (newHover != this.context.DragPortTarget.DiagramObject))
                    {
                        // new target found
                        newTarget = new DiagramObjectTarget(info);
                        this.context.UpdateDragingPort(newTarget, e.Location);
                        break;
                    }
                }
                if (newTarget == null)
                {
                    this.context.UpdateDragingPort(e.Location);
                }
            }
        }

        protected override void OnMouseDown(MouseEventArgs e)
        {
            base.OnMouseDown(e);
            this.diagram.ClearErrorMessage();

            if ((e.Button == MouseButtons.Left) &&
                (this.context.DiagramState == DiagramState.HoverEntity))
            {
                if (this.context.HoverInfo is HoverDiagramObjectInfo)
                {
                    IDiagramObject obj = ((HoverDiagramObjectInfo)this.context.HoverInfo).DiagramObject;
                    DiagramObjectInfo objInfo = Array.Find<DiagramObjectInfo>(
                        this.context.DiagramObjectInfos,
                        delegate(DiagramObjectInfo info)
                        {
                            return ((info != null) && (info.DiagramObject == obj));
                        });
                    if (objInfo != null)
                    {
                        this.context.StartMovingObject(
                            objInfo,
                            new Size(obj.Location.X - e.X, obj.Location.Y - e.Y));
                    }
                }
                else if (this.context.HoverInfo is HoverDiagramObjectPortInfo)
                {
                    HoverDiagramObjectPortInfo port = (HoverDiagramObjectPortInfo)this.context.HoverInfo;
                    this.context.StartDragingPort(port);
                    this.context.UpdateDragingPort(e.Location);
                }
            }
            else if ((e.Button == MouseButtons.Right) &&
                (this.context.DiagramState == DiagramState.HoverEntity))
            {
                if (this.context.HoverInfo.InnerEntity is IDiagramHasActions)
                {
                    IDiagramHasActions actionContainer = 
                        (IDiagramHasActions)this.context.HoverInfo.InnerEntity;
                    ContextMenu menu = new ContextMenu();
                    EventHandler onClick =
                        delegate(Object sender, EventArgs args)
                        {
                            if ((sender is MenuItem) &&
                                (((MenuItem)sender).Tag is KeyValuePair<String, IDiagramHasActions>))
                            {
                                KeyValuePair<String, IDiagramHasActions> pair =
                                    (KeyValuePair<String, IDiagramHasActions>)((MenuItem)sender).Tag;
                                if ((pair.Value != null) && (String.IsNullOrEmpty(pair.Key) == false))
                                    pair.Value.PerformAction(pair.Key);
                            }
                        };
                    foreach (KeyValuePair<String, IDiagramAction> pair in actionContainer.GetActions())
                    {
                        if (pair.Value.Text == "-")
                            menu.MenuItems.Add("-");
                        else
                        {
                            MenuItem item = new MenuItem();
                            item.Text = pair.Value.Text;
                            item.Enabled = pair.Value.Enabled;
                            item.Tag = new KeyValuePair<String, IDiagramHasActions>(
                                pair.Key,
                                actionContainer);
                            item.Click += onClick;
                            menu.MenuItems.Add(item);
                        }
                    }
                    menu.Show(this, e.Location);
                }
            }
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            base.OnMouseUp(e);

            if (this.context.DiagramState == DiagramState.MoveObject)
            {
                IDiagramObject obj = this.context.MovingObject.DiagramObject;
                HoverDiagramObjectInfo newHover =
                    new HoverDiagramObjectInfo(this.context.MovingObject.DiagramObject);
                this.context.ClearState();
                this.context.SetHoveringEntity(newHover);
            }
            else if (this.context.DiagramState == DiagramState.DragPort)
            {
                if ((this.context.DragPortTarget != null) &&
                    (this.context.DragPortTarget.DiagramObject != null) &&
                    (this.diagram.ContainsObject(this.context.DragPortTarget.DiagramObject) == true))
                {
                    this.diagram.NewLinkFromUI(
                        this.context.DragPort.DiagramObject,
                        this.context.DragPort.Port,
                        this.context.DragPort.PortAlign,
                        this.context.DragPortTarget.DiagramObject,
                        this.context.DragPortTarget.DiagramObjectEdge);
                }
                this.context.ClearState();
            }
        }

        private void DiagramView_DragEnter(object sender, DragEventArgs e)
        {
            if ((e.Data != null) && (e.Data.GetDataPresent(typeof(DiagramViewDropItem)) == true))
            {
                e.Effect = DragDropEffects.All;
            }
            else
            {
                e.Effect = DragDropEffects.None;
            }
        }

        private void DiagramView_DragDrop(object sender, DragEventArgs e)
        {
            if ((e.Data != null) && (e.Data.GetDataPresent(typeof(DiagramViewDropItem)) == true))
            {
                DiagramViewDropItem item = (DiagramViewDropItem)
                    e.Data.GetData(typeof(DiagramViewDropItem));
                this.diagram.SomethingDropedOnUI(item.DropInfo,
                    this.PointToClient(new Point(e.X, e.Y)));
            }   
        }
    }
}