using System;
using System.Drawing;
using System.Collections.Generic;
using System.Text;

using Softix.Drawing.Diagram.Core;
using Softix.Drawing.Diagram.Core.Events;

namespace Softix.Drawing.Diagram.Internals
{
    internal sealed class DiagramContext
    {
        private DiagramView view = null;
        private IDiagram diagram = null;
        private Dictionary<IDiagramObject, DiagramObjectInfo> infos = null;
        // private Dictionary<IDiagramObjectLink, DiagramObjectLinkInfo> links = null;

        public DiagramContext(DiagramView view, IDiagram diagram)
        {
            this.view = view;
            this.diagram = (diagram == null) ? view.Diagram : diagram;
            this.infos = new Dictionary<IDiagramObject, DiagramObjectInfo>();
            this.ClearState();

            // set events
            this.diagram.DiagramObjectAdded += new DiagramObjectEventHandler(HandleDiagramObjectAdded);
            this.diagram.DiagramObjectRemoved += new DiagramObjectEventHandler(HandleDiagramObjectRemoved);
            this.diagram.DiagramObjectLinkAdded += new DiagramObjectLinkEventHandler(HandleDiagramObjectLinkAdded);

            foreach (IDiagramObject o in this.diagram.DiagramObjects)
            {
                infos.Add(o, new DiagramObjectInfo(o));
                o.ShouldRedraw += new EventHandler(ObjectShouldRedraw);
            }
            this.InvalidateView();
        }

        void HandleDiagramObjectRemoved(IDiagram sender, IDiagramObject obj)
        {
            if (infos.ContainsKey(obj) == true)
            {
                infos.Remove(obj);
                this.InvalidateView();
            }
        }

        void HandleDiagramObjectAdded(IDiagram sender, IDiagramObject obj)
        {
            infos.Add(obj, new DiagramObjectInfo(obj));
            obj.ShouldRedraw += new EventHandler(ObjectShouldRedraw);
            this.InvalidateView();
        }

        void ObjectShouldRedraw(object sender, EventArgs e)
        {
            if (sender is IDiagramObject)
            {
                IDiagramObject diagramObject = (IDiagramObject)sender;
                if (this.infos.ContainsKey(diagramObject) == true)
                {
                    this.infos[diagramObject].Dirty = true;
                    this.InvalidateView();
                }
            }
        }

        void HandleDiagramObjectLinkAdded(IDiagram sender, IDiagramObjectLink link)
        {
            // infos.Add(obj, new DiagramObjectInfo(obj));
            this.InvalidateView();
        }

        internal DiagramView View { get { return this.view; } }
        internal DiagramObjectInfo[] DiagramObjectInfos
        {
            get
            {
                DiagramObjectInfo[] infos = new DiagramObjectInfo[this.infos.Count];
                this.infos.Values.CopyTo(infos, 0);
                return infos;
            }
        }

        internal void InvalidateView()
        {
            if (this.view != null)
                this.view.Invalidate();
        }

        #region public DiagramState DiagramState { get; set; }

        private DiagramState state;

        public DiagramState DiagramState
        {
            get { return state; }
            set { state = value; }
        }

        #endregion

        //////////
        // None state

        internal void ClearState()
        {
            if (this.state != DiagramState.None)
            {
                this.state = DiagramState.None;
                this.hoverInfo = null;
                this.dragPort = null;
                this.dragPortTarget = null;
                this.dragPortTargetPoint = Point.Empty;
                this.movingObject = null;
                this.movingObjectOffset = Size.Empty;

                this.InvalidateView();
            }
        }

        //////////
        // HoverEntity state

        #region public IHoverObjectInfo HoverInfo { get; }

        private IHoverObjectInfo hoverInfo;

        public IHoverObjectInfo HoverInfo
        {
            get
            {
                return (hoverInfo == null)
                    ? new HoverDummyInfo()
                    : hoverInfo;
            }
            // set { hoverInfo = value; }
        }

        #endregion

        internal void SetHoveringEntity(IHoverObjectInfo hoverInfo)
        {
            if ((this.state != DiagramState.None) &&
                (this.state != DiagramState.HoverEntity))
                throw new Exception("Incorrect diagram state.");

            this.state = DiagramState.HoverEntity;
            if ((this.hoverInfo == null) ||
                (this.hoverInfo.Equals(hoverInfo) == false))
            {
                this.hoverInfo = hoverInfo;
                this.InvalidateView();
            }
        }

        //////////
        // MoveObject state

        #region public DiagramObjectInfo MovingObject { get; }

        private DiagramObjectInfo movingObject;

        public DiagramObjectInfo MovingObject
        {
            get { return movingObject; }
            // set { movingObject = value; }
        }

        #endregion
        #region public Size MovingObjectOffset { get; }

        private Size movingObjectOffset;

        public Size MovingObjectOffset
        {
            get { return movingObjectOffset; }
        }

        #endregion
            
        internal void StartMovingObject(DiagramObjectInfo movingObject, Size offset)
        {
            if (this.state != DiagramState.HoverEntity)
                throw new Exception("Incorrect diagram state.");

            this.state = DiagramState.MoveObject;
            this.movingObject = movingObject;
            this.movingObjectOffset = offset;
            this.InvalidateView();
        }

        internal void UpdateMovingObject(Point newLocation)
        {
            newLocation.Offset(this.movingObjectOffset.Width, this.movingObjectOffset.Height);
            this.movingObject.DiagramObject.Location = newLocation;
            this.InvalidateView();
        }

        //////////
        // DragPort state

        #region public HoverDiagramObjectPortInfo DragPort { get; }

        private HoverDiagramObjectPortInfo dragPort;

        public HoverDiagramObjectPortInfo DragPort
        {
            get { return dragPort; }
            // set { dragPort = value; }
        }

        #endregion
        #region public DiagramObjectTarget DragPortTarget { get; }

        private DiagramObjectTarget dragPortTarget;

        public DiagramObjectTarget DragPortTarget
        {
            get { return dragPortTarget; }
            // set { dragPortTarget = value; }
        }

        #endregion
        #region public Point DragPortTargetPoint { get; }

        private Point dragPortTargetPoint;

        public Point DragPortTargetPoint
        {
            get { return dragPortTargetPoint; }
            // set { dragPortTargetPoint = value; }
        }

        #endregion

        internal void StartDragingPort(HoverDiagramObjectPortInfo portInfo)
        {
            this.state = DiagramState.DragPort;

            this.dragPort = portInfo;
            this.dragPortTarget = null;
            this.dragPortTargetPoint = Point.Empty;
        }

        internal void UpdateDragingPort(DiagramObjectTarget target, Point mouseLocation)
        {
            if (target != this.dragPortTarget)
                this.dragPortTarget = target;
            this.dragPortTarget.RefreshInfo(mouseLocation);
            this.dragPortTargetPoint = target.TargetPoint;
            this.InvalidateView();
        }
        
        internal void UpdateDragingPort(Point targetLocation)
        {
            this.dragPortTarget = null;
            this.dragPortTargetPoint = targetLocation;
            this.InvalidateView();
        }

        //////////
        // draw things

        internal void DrawObjects(Graphics g)
        {
            foreach (DiagramObjectInfo info in this.infos.Values)
                info.Draw(g, this);
        }
        
        internal void DrawSpecialUI(Graphics g)
        {
            if (this.state == DiagramState.DragPort)
            {
                if ((this.dragPortTarget != null) && (dragPortTarget.DiagramObject != null))
                {
                    #region draw target object mask

                    DiagramObjectInfo objectInfo = this.infos[this.dragPortTarget.DiagramObject];
                    using(Brush brush = new SolidBrush(Color.FromArgb(100, Color.Snow)))
                    using(Pen pen = new Pen(Color.Goldenrod))
                    {
                        // pen.DashStyle = System.Drawing.Drawing2D.DashStyle.Dot;
                        pen.DashPattern = new float[] { 1.0f, 3.0f };
                        g.FillRectangle(brush, objectInfo.BoundRect);
                        g.DrawLine(pen,
                            objectInfo.BoundRect.Left, objectInfo.BoundRect.Top,
                            objectInfo.BoundRect.Right, objectInfo.BoundRect.Bottom);
                        g.DrawLine(pen,
                            objectInfo.BoundRect.Left, objectInfo.BoundRect.Bottom,
                            objectInfo.BoundRect.Right, objectInfo.BoundRect.Top);
                    }

                    #endregion
                }

                if (this.dragPort is HoverDiagramObjectPortInfo)
                {
                    HoverDiagramObjectPortInfo port = (HoverDiagramObjectPortInfo)this.dragPort;
                    DiagramObjectInfo objectInfo = this.infos[port.DiagramObject];

                    #region draw drag port line

                    if ((objectInfo != null) && (objectInfo.Ports.ContainsKey(port.Port) == true))
                    {
                        DiagramObjectPortInfo portInfo = objectInfo.Ports[port.Port];
                        using (Pen pen = new Pen(Color.DarkGreen))
                        {
                            pen.CustomEndCap = new System.Drawing.Drawing2D.AdjustableArrowCap(5, 5);
                            Point startPoint = Point.Empty;
                            if (this.dragPort.PortAlign == DiagramObjectPortAlign.Left)
                                startPoint = new Point(
                                    portInfo.LeftPort.Left - 1,
                                    portInfo.LeftPort.Top + portInfo.LeftPort.Height / 2);
                            else
                                startPoint = new Point(
                                    portInfo.RightPort.Right + 1,
                                    portInfo.RightPort.Top + portInfo.RightPort.Height / 2);
                            g.DrawLine(pen, startPoint, this.dragPortTargetPoint);
                        }
                    }

                    #endregion
                }
            }
        }

        internal void DrawLinks(Graphics g)
        {
            IDiagramObjectLink[] links = this.diagram.DiagramObjectLinks;
            using (Pen linkPen = new Pen(Color.SteelBlue))
            {
                linkPen.CustomEndCap = new System.Drawing.Drawing2D.AdjustableArrowCap(5, 5);
                linkPen.StartCap = System.Drawing.Drawing2D.LineCap.DiamondAnchor;
                foreach (IDiagramObjectLink link in links)
                {
                    if ((link.Start == null) || (this.infos.ContainsKey(link.Start) == false))
                        continue;
                    DiagramObjectInfo start = this.infos[link.Start];

                    if ((link.StartPort == null) || (start.Ports.ContainsKey(link.StartPort) == false))
                        return;
                    DiagramObjectPortInfo port = start.Ports[link.StartPort];
                    Rectangle portRect = port.BoundRect;

                    if ((link.End == null) || (this.infos.ContainsKey(link.End) == false))
                        continue;
                    DiagramObjectInfo end = this.infos[link.End];

                    Point pStart = new Point(
                        (link.StartPortAlign == DiagramObjectPortAlign.Left) ? portRect.Left - 6 : portRect.Right + 6,
                        portRect.Top + (portRect.Bottom - portRect.Top) / 2);
                    Point pEnd = Point.Empty;
                    switch (link.EndEdge)
                    {
                        case DiagramObjectEdge.Left:
                            pEnd = new Point(
                                end.BoundRect.Left - 1,
                                end.ImageRect.Top + (end.ImageRect.Bottom - end.ImageRect.Top) / 2);
                            break;
                        case DiagramObjectEdge.Top:
                            pEnd = new Point(
                                end.ImageRect.Left + (end.ImageRect.Right - end.ImageRect.Left) / 2,
                                end.BoundRect.Top - 1);
                            break;
                        case DiagramObjectEdge.Right:
                            pEnd = new Point(
                                end.BoundRect.Right + 1,
                                end.ImageRect.Top + (end.ImageRect.Bottom - end.ImageRect.Top) / 2);
                            break;
                        case DiagramObjectEdge.Bottom:
                            pEnd = new Point(
                                end.ImageRect.Left + (end.ImageRect.Right - end.ImageRect.Left) / 2,
                                end.BoundRect.Bottom + 1);
                            break;
                    }
                    g.DrawLine(linkPen, pStart, pEnd);
                }
            }
        }
    }
}
