using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;

using Softix.Drawing.Diagram.Core;

using DiagramObjectInfoPortListItem = System.Collections.Generic.KeyValuePair<
    Softix.Drawing.Diagram.Core.IDiagramObjectPort,
    Softix.Drawing.Diagram.Internals.DiagramObjectPortInfo>;

namespace Softix.Drawing.Diagram.Internals
{
    internal sealed class DiagramObjectInfo
    {
        internal DiagramObjectInfo(IDiagramObject obj)
        {
            this.diagramObject = obj;
            this.dirty = true;
        }

        #region internal IDiagramObject DiagramObject { get; }

        private IDiagramObject diagramObject;

        internal IDiagramObject DiagramObject
        {
            get { return diagramObject; }
            // set { diagramObject = value; }
        }

        #endregion
        #region public bool Dirty { get; set; }

        private bool dirty;

        public bool Dirty
        {
            get { return dirty; }
            set { dirty = value; }
        }

        #endregion

        internal void RefreshInfo(DiagramView view, Graphics g)
        {
            StringFormat sf = new StringFormat();
            sf.FormatFlags = StringFormatFlags.FitBlackBox;
            sf.Alignment = StringAlignment.Center;
            sf.LineAlignment = StringAlignment.Center;
            sf.Trimming = StringTrimming.Word;
            this.stringFormat = sf;

            Size nameSize = Size.Ceiling(g.MeasureString(this.diagramObject.Text, view.Font, 100, sf));

            Rectangle rect = new Rectangle();
            rect.Width = Math.Max(nameSize.Width, 32) + 6;
            rect.Height = 32 + 3 + (nameSize.Height + 6);
            rect.X = this.diagramObject.Location.X - rect.Width / 2;
            rect.Y = this.diagramObject.Location.Y - rect.Height / 2;
            this.boundRect = rect;

            // region name rect
            Rectangle nameRect = new Rectangle(0, 0, nameSize.Width, nameSize.Height);
            nameRect.X = rect.Left + (rect.Width - nameSize.Width) / 2;
            nameRect.Y = rect.Top + 3;
            this.nameRect = nameRect;

            // region image
            Rectangle imageRect = new Rectangle(0, 0, 32, 32);
            imageRect.X = rect.Left + (rect.Width - 32) / 2;
            imageRect.Y = nameRect.Bottom + 3;
            this.imageRect = imageRect;

            // set port bounds
            DiagramObjectPortInfoList list = new DiagramObjectPortInfoList();
            IDiagramObjectPort[] ports = this.diagramObject.Ports;
            if ((ports != null) || (ports.Length > 0))
            {
                foreach (IDiagramObjectPort port in ports)
                {
                    DiagramObjectPortInfo info = new DiagramObjectPortInfo(
                        this, port, view, g);
                    list.Add(port, info);
                }

                foreach (DiagramObjectPortInfo portInfo in list.Values)
                {
                    Rectangle tmpRect = portInfo.BoundRect;
                    tmpRect.X = this.boundRect.Left;
                    tmpRect.Width = this.boundRect.Width;
                    portInfo.BoundRect = tmpRect;
                }
            }
            this.ports = list;

            this.dirty = false;
        }

        #region internal StringFormat StringFormat { get; }

        private StringFormat stringFormat;

        internal StringFormat StringFormat
        {
            get { return stringFormat; }
            // set { stringFormat = value; }
        }

        #endregion
        #region internal Font Font { get; }

        private Font font = new Font("Tahoma", 8.25f, FontStyle.Regular, GraphicsUnit.Point);

        internal Font Font
        {
            get { return font; }
            // set { font = value; }
        }

        #endregion
        #region internal Rectangle BoundRect { get; set; }

        private Rectangle boundRect;

        internal Rectangle BoundRect
        {
            get { return boundRect; }
            set
            {
                boundRect = value;
                boundRect.X = this.diagramObject.Location.X - boundRect.Width / 2;
                if (this.nameRect.IsEmpty == false)
                {
                    nameRect.X = boundRect.Left + (boundRect.Width - nameRect.Width) / 2;
                    nameRect.Y = boundRect.Top + 3;
                }
                if (this.imageRect.IsEmpty == false)
                {
                    imageRect.X = boundRect.Left + (boundRect.Width - imageRect.Width) / 2;
                    imageRect.Y = nameRect.Bottom + 3;
                }
            }
        }

        #endregion
        #region internal Rectangle NameRect { get; }

        private Rectangle nameRect = Rectangle.Empty;

        internal Rectangle NameRect
        {
            get { return nameRect; }
            // set { nameRect = value; }
        }

        #endregion
        #region internal Rectangle ImageRect { get; }

        private Rectangle imageRect = Rectangle.Empty;

        internal Rectangle ImageRect
        {
            get { return imageRect; }
            // set { imageRect = value; }
        }

        #endregion

        #region public DiagramObjectPortInfoList Ports { get; }

        private DiagramObjectPortInfoList ports;

        public DiagramObjectPortInfoList Ports
        {
            get { return ports; }
            // set { ports = value; }
        }

        #endregion

        internal IHoverObjectInfo TestHover(Point p)
        {
            return this.TestHover(p, true);
        }

        internal IHoverObjectInfo TestHover(Point p, bool testPorts)
        {
            Rectangle bigRect = this.boundRect;
            bigRect.Inflate(20, 20);
            if (bigRect.Contains(p) == false)
                return null;

            if ((testPorts == true) && (this.ports != null) && (this.ports.Count > 0))
            {
                foreach (DiagramObjectPortInfo port in this.ports.Values)
                {
                    if (port.LeftPort.Contains(p) == true)
                        return new HoverDiagramObjectPortInfo(
                            this.diagramObject,
                            port.Port,
                            DiagramObjectPortAlign.Left);
                    if (port.RightPort.Contains(p) == true)
                        return new HoverDiagramObjectPortInfo(
                            this.diagramObject,
                            port.Port,
                            DiagramObjectPortAlign.Right);
                }
            }

            if (this.boundRect.Contains(p) == true)
                return new HoverDiagramObjectInfo(this.DiagramObject);

            return null;
        }

        internal void Draw(Graphics g, DiagramContext context)
        {
            if (this.dirty == true)
                this.RefreshInfo(context.View, g);

            #region background, text & image

            if ((context.DiagramState == DiagramState.MoveObject) &&
                (context.MovingObject.DiagramObject == this.diagramObject))
            {
                Rectangle tmpRect = this.BoundRect;
                tmpRect.Offset(3, 3);
                using (SolidBrush br = new SolidBrush(Color.FromArgb(50, Color.Black)))
                {
                    g.FillRectangle(br, tmpRect);
                }
                g.FillRectangle(Brushes.BlanchedAlmond, this.BoundRect);
                g.DrawRectangle(Pens.Black, this.BoundRect);
            }
            else
            {
                using (Brush bgBrush = new System.Drawing.Drawing2D.LinearGradientBrush(
                    this.BoundRect, Color.WhiteSmoke, Color.BlanchedAlmond,
                    System.Drawing.Drawing2D.LinearGradientMode.Vertical))
                {
                    g.FillRectangle(bgBrush, this.BoundRect);
                }
                if ((context.DiagramState == DiagramState.HoverEntity) &&
                    (context.HoverInfo.IsMatch(this.diagramObject) == true))
                {
                    g.DrawRectangle(Pens.Blue, this.BoundRect);
                }
                else
                {
                    g.DrawRectangle(Pens.Black, this.BoundRect);
                }
            }
            g.DrawString(this.DiagramObject.Text, this.Font, Brushes.Black, this.NameRect, this.StringFormat);
            using (Image img = Softix.Drawing.Diagram.Resouces.Resources.DefaultComponentImage)
            {
                g.DrawImage(img, this.ImageRect);
            }
            g.DrawLine(Pens.Silver,
                this.BoundRect.Left + 2, this.ImageRect.Top - 2,
                this.BoundRect.Right - 2, this.ImageRect.Top - 2);

            #endregion

            // draw ports
            foreach (DiagramObjectInfoPortListItem entry in this.Ports)
            {
                Rectangle portRect = entry.Value.BoundRect;

                if ((entry.Key.BackgroundColor != Color.Empty) &&
                    (entry.Key.BackgroundColor != Color.Transparent))
                {
                    using (Brush br = new SolidBrush(entry.Key.BackgroundColor))
                    {
                        g.FillRectangle(br, entry.Value.BoundRect);
                    }
                }
                g.DrawString(entry.Key.Text, this.Font, Brushes.DarkGreen, portRect, this.StringFormat);
                g.DrawLine(Pens.Silver,
                    this.BoundRect.Left + 2, portRect.Top - 1,
                    this.BoundRect.Right - 2, portRect.Top - 1);

                #region draw ports

                Rectangle tmpPortRect = Rectangle.Empty;

                // draw left port
                tmpPortRect = new Rectangle(
                        portRect.Left - 5, portRect.Top + (portRect.Height - 7) / 2,
                        7, 7);
                using (Brush br = new System.Drawing.Drawing2D.LinearGradientBrush(
                    entry.Value.LeftPort, Color.Silver, Color.White,
                    System.Drawing.Drawing2D.LinearGradientMode.ForwardDiagonal))
                {
                    g.FillRectangle(br, entry.Value.LeftPort);
                }
                Pen leftPortColor = Pens.DarkGray;
                if ((context.DiagramState == DiagramState.HoverEntity) &&
                    (context.HoverInfo.IsMatch(entry.Key, DiagramObjectPortAlign.Left) == true))
                {
                    leftPortColor = Pens.Blue;
                }
                g.DrawRectangle(leftPortColor, entry.Value.LeftPort);

                // draw right port
                tmpPortRect = new Rectangle(
                    portRect.Right - 2, portRect.Top + (portRect.Height - 7) / 2,
                    7, 7);
                using (Brush br = new System.Drawing.Drawing2D.LinearGradientBrush(
                    entry.Value.RightPort, Color.Silver, Color.White,
                    System.Drawing.Drawing2D.LinearGradientMode.ForwardDiagonal))
                {
                    g.FillRectangle(br, entry.Value.RightPort);
                }
                Pen rightPortColor = Pens.DarkGray;
                if ((context.DiagramState == DiagramState.HoverEntity) &&
                    (context.HoverInfo.IsMatch(entry.Key, DiagramObjectPortAlign.Right) == true))
                {
                    rightPortColor = Pens.Blue;
                }
                g.DrawRectangle(rightPortColor, entry.Value.RightPort);

                #endregion
            }

            // draw artifacts
            int leftOffset = 0, rightOffset = 0, topOffset = 0, bottomOffset = 0;
            Size artifactSize = new Size(18, 18);
            if (this.diagramObject.Artifacts != null)
            {
                foreach (IDiagramArtifact artifact in this.diagramObject.Artifacts)
                {
                    if (artifact.Image == null)
                        continue;

                    Rectangle artifactRect = new Rectangle(
                        0, 0,
                        artifactSize.Width, artifactSize.Height);
                    switch(artifact.Edge)
                    {
                        case DiagramObjectEdge.Left:
                            artifactRect.Location = new Point(
                                this.boundRect.Left - (artifactSize.Width + 1),
                                this.boundRect.Top + leftOffset);
                            leftOffset += artifactSize.Height + 1;
                            break;
                        case DiagramObjectEdge.Top:
                            artifactRect.Location = new Point(
                                this.boundRect.Left + topOffset,
                                this.boundRect.Top - (artifactSize.Height + 1));
                            topOffset += artifactSize.Width + 1;
                            break;
                        case DiagramObjectEdge.Right:
                            artifactRect.Location = new Point(
                                this.boundRect.Right + 1,
                                this.boundRect.Top + rightOffset);
                            rightOffset += artifactSize.Height + 1;
                            break;
                        case DiagramObjectEdge.Bottom:
                            artifactRect.Location = new Point(
                                this.boundRect.Left + bottomOffset,
                                this.boundRect.Bottom + 1);
                            bottomOffset += artifactSize.Width + 1;
                            break;
                    }
                    g.DrawImage(artifact.Image, artifactRect);
                }
            }
        }
    }
}
