using System;
using System.Drawing;
using System.Collections.Generic;
using System.Text;
using System.Xml.Serialization;

using Softix.Drawing.Diagram.Core;
using Softix.Drawing.Diagram.Core.Events;

namespace Softix.Drawing.Diagram.Simple
{
    public class SimpleDiagram : IDiagram
    {
        private List<IDiagramEntity> entities;

        public SimpleDiagram()
        {
            this.CheckEntityList();
        }

        #region DiagramObjects, DiagramObjectLinks

        private void CheckEntityList()
        {
            if (this.entities == null)
                this.entities = new List<IDiagramEntity>();
        }

        public IDiagramObject[] DiagramObjects
        {
            get
            {
                if (this.entities == null)
                    return new IDiagramObject[0];

                return this.entities.FindAll(
                    delegate(IDiagramEntity entity)
                    {
                        return (entity is IDiagramObject);
                    }).ConvertAll<IDiagramObject>(
                    delegate(IDiagramEntity entity)
                    {
                        return (entity as IDiagramObject);
                    }).ToArray();
            }
        }

        public IDiagramObjectLink[] DiagramObjectLinks
        {
            get
            {
                if (this.entities == null)
                    return new IDiagramObjectLink[0];

                return this.entities.FindAll(
                    delegate(IDiagramEntity entity)
                    {
                        return (entity is IDiagramObjectLink);
                    }).ConvertAll<IDiagramObjectLink>(
                    delegate(IDiagramEntity entity)
                    {
                        return (entity as IDiagramObjectLink);
                    }).ToArray();
            }
        }

        public bool ContainsObject(IDiagramObject obj)
        {
            this.CheckEntityList();
            return this.entities.Contains(obj);
        }

        public bool ContainsObjectLink(IDiagramObjectLink link)
        {
            this.CheckEntityList();
            return this.entities.Contains(link);
        }

        #endregion

        #region add & remove diagram object

        public virtual void AddDiagramObject(IDiagramObject diagramObject)
        {
            this.CheckEntityList();
            if (this.entities.Contains(diagramObject) == false)
            {
                this.entities.Add(diagramObject);
                diagramObject.ShouldRedraw += new EventHandler(DiagramObject_ShouldRedraw);
                this.OnDiagramObjectAdded(diagramObject);
            }
        }

        void DiagramObject_ShouldRedraw(object sender, EventArgs e)
        {
            this.OnShouldRedraw();
        }

        public virtual void RemoveDiagramObject(IDiagramObject diagramObject)
        {
            this.CheckEntityList();
            if (this.entities.Contains(diagramObject) == true)
            {
                diagramObject.ShouldRedraw -= new EventHandler(DiagramObject_ShouldRedraw);
                foreach (IDiagramObjectLink link in this.DiagramObjectLinks)
                {
                    if ((link.Start == diagramObject) || (link.End == diagramObject))
                        this.RemoveDiagramObjectLink(link);
                }

                this.entities.Remove(diagramObject);
                this.OnDiagramObjectRemoved(diagramObject);
            }
        }

        public event DiagramObjectEventHandler DiagramObjectAdded;
        public event DiagramObjectEventHandler DiagramObjectRemoved;

        protected virtual void OnDiagramObjectAdded(IDiagramObject diagramObject)
        {
            if (this.DiagramObjectAdded != null)
                this.DiagramObjectAdded(this, diagramObject);
            this.OnShouldRedraw();
        }

        protected virtual void OnDiagramObjectRemoved(IDiagramObject diagramObject)
        {
            if (this.DiagramObjectRemoved != null)
                this.DiagramObjectRemoved(this, diagramObject);
            this.OnShouldRedraw();
        }

        #endregion

        #region add & remove diagram object link

        public void AddDiagramObjectLink(IDiagramObjectLink diagramLink)
        {
            this.CheckEntityList();
            if (this.entities.Contains(diagramLink) == false)
            {
                if (this.entities.Contains(diagramLink.Start) == false)
                    throw new Exception("Link's start object must be on this diagram.");
                if (this.entities.Contains(diagramLink.End) == false)
                    throw new Exception("Link's end object must be on this diagram.");

                this.entities.Add(diagramLink);
                this.OnDiagramObjectLinkAdded(diagramLink);
            }
        }

        public void RemoveDiagramObjectLink(IDiagramObjectLink diagramLink)
        {
            this.CheckEntityList();
            if (this.entities.Contains(diagramLink) == true)
            {
                this.entities.Remove(diagramLink);
                this.OnDiagramObjectLinkRemoved(diagramLink);
            }
        }

        public event DiagramObjectLinkEventHandler DiagramObjectLinkAdded;
        public event DiagramObjectLinkEventHandler DiagramObjectLinkRemoved;

        protected virtual void OnDiagramObjectLinkAdded(IDiagramObjectLink diagramLink)
        {
            if (this.DiagramObjectLinkAdded != null)
                this.DiagramObjectLinkAdded(this, diagramLink);
            this.OnShouldRedraw();
        }

        protected virtual void OnDiagramObjectLinkRemoved(IDiagramObjectLink diagramLink)
        {
            if (this.DiagramObjectLinkRemoved != null)
                this.DiagramObjectLinkRemoved(this, diagramLink);
            this.OnShouldRedraw();
        }

        #endregion

        #region add diagram entities

        public void AddDiagramEntities(IDiagramEntity[] entities)
        {
            this.SuspendRedraw();
            try
            {
                if ((entities != null) && (entities.Length > 0))
                {
                    List<IDiagramObjectLink> links = new List<IDiagramObjectLink>();

                    foreach (IDiagramEntity entity in entities)
                    {
                        if (entity is IDiagramObjectLink)
                        {
                            // we add links after we add all objects
                            links.Add((IDiagramObjectLink)entity);
                        }
                        else if (entity is IDiagramObject)
                        {
                            // add diagram object
                            IDiagramObject diagramObject = (IDiagramObject)entity;
                            this.AddDiagramObject(diagramObject);
                        }
                        else
                            throw new Exception("Unknown diagram entity type.");
                    }

                    if (links.Count > 0)
                    {
                        // add diagram links
                        foreach (IDiagramObjectLink link in links)
                        {
                            this.AddDiagramObjectLink(link);
                        }
                    }
                }
            }
            finally
            {
                this.ResumeRedraw(true);
            }
        }

        #endregion

        #region public IDiagramEntity SelectedDiagramEntity { get; set; }

        private IDiagramEntity selectedDiagramEntity = null;

        public IDiagramEntity SelectedDiagramEntity
        {
            get { return this.selectedDiagramEntity; }
            set
            {
                if (this.selectedDiagramEntity != value)
                {
                    this.selectedDiagramEntity = value;
                    this.OnSelectedEntityChanged(value);
                }
            }
        }

        public event DiagramEntityEventHandler SelectedEntityChanged;

        protected virtual void OnSelectedEntityChanged(IDiagramEntity entity)
        {
            if (this.SelectedEntityChanged != null)
                this.SelectedEntityChanged(this, entity);
        }

        #endregion

        #region error message

        private string errorMessage = String.Empty;

        public String ErrorMessage
        {
            get { return this.errorMessage; }
            protected set
            {
                if (this.errorMessage != value)
                {
                    this.errorMessage = value;
                    this.OnShouldRedraw();
                }
            }
        }

        public void ClearErrorMessage() { this.ErrorMessage = ""; }

        #endregion

        #region should redraw event

        public event EventHandler ShouldRedraw;

        private bool fireShouldRedraw = true;

        protected virtual void OnShouldRedraw()
        {
            if ((this.fireShouldRedraw == true) && (this.ShouldRedraw != null))
                this.ShouldRedraw(this, EventArgs.Empty);
        }

        protected void SuspendRedraw() { this.fireShouldRedraw = false; }

        protected void ResumeRedraw(bool forceRedraw)
        {
            this.fireShouldRedraw = true;
            if (forceRedraw == true)
                this.OnShouldRedraw();
        }

        #endregion

        public virtual void NewLinkFromUI(
            IDiagramObject source,
            IDiagramObjectPort sourcePort,
            DiagramObjectPortAlign sourcePortAlign,
            IDiagramObject targetObject,
            DiagramObjectEdge targetEdge)
        {
            SimpleDiagramLink link = new SimpleDiagramLink();
            link.Start = source;
            link.StartPort = sourcePort;
            link.StartPortAlign = sourcePortAlign;
            link.End = targetObject;
            link.EndEdge = targetEdge;

            this.AddDiagramObjectLink(link);
        }

        public virtual void SomethingDropedOnUI(Object dropedItem, Point location)
        {
            SimpleDiagramObject obj = new SimpleDiagramObject(
                dropedItem.GetType().Name,
                dropedItem.ToString(),
                location);
            obj.AddArtifact("warning", SimpleDiagramArtifact.WarningArtifact);

            this.AddDiagramObject(obj);
        }
    }
}
