using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;

using Softix.Orchestration.Core.Data;
using Softix.Orchestration.Core.Models;
using Softix.Orchestration.Core.Validation;
using Softix.Orchestration.Core.Components.Events;

namespace Softix.Orchestration.Core.Components
{
    [Serializable]
    public abstract class ComponentBase : ValidationBase, ISOComponent
    {
        public ComponentBase()
        {
            this.id = Guid.NewGuid();
        }

        #region public String Name { get; set; }
      
        private String name;

        [Category("2. Settings")]
        [Browsable(true)]
        public String Name
        {
            get { return name;}
            set
            {
                name = value;
                this.OnComponentChanged();
            }
        }

        #endregion
        #region public Guid ID { get; set; }

        private Guid id;

        [ Category("1. Information") ]
        [ Browsable(true) ]
        public Guid ID
        {
            get { return id; }
            protected set { id = value; }
        }

        #endregion

        #region public ISOModel ParentModel { get; set; }

        private ISOModel model;

        [ Browsable(false) ]
        public ISOModel ParentModel
        {
            get { return model; }
            set
            {
                if (model != value)
                {
                    if (model != null)
                    {
                        // remove event handlers from old model
                        model.ComponentLinkAdded -= new Softix.Orchestration.Core.Models.Events.ModelComponentLinkHandler(Model_ComponentLinkAdded);
                        model.ComponentLinkRemoved -= new Softix.Orchestration.Core.Models.Events.ModelComponentLinkHandler(Model_ComponentLinkRemoved);
                        model.ComponentLinkChanged -= new Softix.Orchestration.Core.Models.Events.ModelComponentLinkHandler(Model_ComponentLinkChanged);
                    }

                    // set new value
                    model = value;
                    if (model != null)
                    {
                        // subscribe to new model's events
                        model.ComponentLinkAdded += new Softix.Orchestration.Core.Models.Events.ModelComponentLinkHandler(Model_ComponentLinkAdded);
                        model.ComponentLinkRemoved += new Softix.Orchestration.Core.Models.Events.ModelComponentLinkHandler(Model_ComponentLinkRemoved);
                        model.ComponentLinkChanged += new Softix.Orchestration.Core.Models.Events.ModelComponentLinkHandler(Model_ComponentLinkChanged);
                    }
                }
            }
        }

        void Model_ComponentLinkChanged(ISOModel model, ISOComponentLink link)
        {
            if ((link.SourceComponent.ID == this.id) ||
                (link.TargetComponent.ID == this.id))
            {
                this.OnComponentChanged();
            }
        }

        void Model_ComponentLinkRemoved(ISOModel model, ISOComponentLink link)
        {
            if ((link.SourceComponent.ID == this.id) ||
                (link.TargetComponent.ID == this.id))
            {
                this.OnComponentChanged();
            }
        }

        void Model_ComponentLinkAdded(ISOModel model, ISOComponentLink link)
        {
            if ((link.SourceComponent.ID == this.id) ||
                (link.TargetComponent.ID == this.id))
            {
                this.OnComponentChanged();
            }
        }

        #endregion

        #region ISOComponent outbound paths & path data definitions

        protected List<ISOComponentOutboundPath> paths = new List<ISOComponentOutboundPath>();

        protected virtual void RefreshPaths() { }

        protected Guid AddOutboundPath(String pathName)
        {
            return this.AddOutboundPath(0, pathName);
        }

        protected Guid AddOutboundPath(Int32 index, String pathName)
        {
            ISOComponentOutboundPath path = this.paths.Find(
                delegate(ISOComponentOutboundPath tmpPath)
                {
                    return tmpPath.Name == pathName;
                });

            if (path == null)
            {
                path = new SimpleOutboundPath(index, pathName);
                this.paths.Add(path);
                this.RunValidation();
            }
            else
                path.Index = index;
            this.paths.Sort(
                delegate(ISOComponentOutboundPath a, ISOComponentOutboundPath b)
                {
                    int cmp = a.Index.CompareTo(b.Index);
                    if (cmp == 0)
                        cmp = a.Name.CompareTo(b.Name);
                    return cmp;
                });
            this.OnComponentChanged();
            return path.ID;
        }

        protected void RemoveOutboundPath(Guid pathID)
        {
            ISOComponentOutboundPath path = this.paths.Find(
                delegate(ISOComponentOutboundPath tmpPath)
                {
                    return tmpPath.ID == pathID;
                });
            this.RemoveOutboundPath(path);
        }

        protected void RemoveOutboundPath(String pathName)
        {
            ISOComponentOutboundPath path = this.paths.Find(
                delegate(ISOComponentOutboundPath tmpPath)
                {
                    return tmpPath.Name == pathName;
                });
            this.RemoveOutboundPath(path);
        }

        private void RemoveOutboundPath(ISOComponentOutboundPath path)
        {
            if (path != null)
            {
                this.paths.Remove(path);

                // remove links from model
                // TODO: move this to model
                if (this.ParentModel != null)
                {
                    ISOComponentLink[] links = this.ParentModel.GetOutboundLinksForComponent(
                        this.id,
                        path.ID);
                    if ((links != null) && (links.Length > 0))
                    {
                        foreach (ISOComponentLink link in links)
                        {
                            this.ParentModel.RemoveComponentLink(link);
                        }
                    }
                }
                this.RunValidation();
                this.OnComponentChanged();
            }
        }

        protected Guid GetPathID(String pathName)
        {
            ISOComponentOutboundPath path = this.paths.Find(
                delegate(ISOComponentOutboundPath tmpPath)
                {
                    return tmpPath.Name == pathName;
                });
            return (path == null) ? Guid.Empty : path.ID;
        }

        [ Browsable(false) ]
        public ISOComponentOutboundPath[] OutboundPaths
        {
            get
            {
                this.RefreshPaths();
                return paths.ToArray();
            }
        }

        protected virtual Variable[] GetDataDefinitionsForPath(String pathName)
        {
            return new Variable[0];
        }

        public virtual Variable[] GetDataDefinitionsForPath(Guid pathID)
        {
            ISOComponentOutboundPath path = this.paths.Find(
                delegate(ISOComponentOutboundPath tmpPath)
                {
                    return (tmpPath.ID == pathID);
                });
            if (path != null)
                return this.GetDataDefinitionsForPath(path.Name);
            throw new InvalidOperationException("Path with id '" + pathID + "' not found.");
        }

        public virtual Variable[] GetDataDefinitionsForPath(ISOComponentOutboundPath path)
        {
            if (this.paths.Contains(path) == false)
                throw new InvalidOperationException("Path with id '" + path.ID + "' not found.");
            return this.GetDataDefinitionsForPath(path.Name);
        }

        #endregion

        #region ISOComponent input data

        [Browsable(false)]
        public virtual Variable[] InputDataDefinitions
        {
            get { return new Variable[0]; }
        }

        #region public VariableMappingList InputDataMappings { get; set; }

        private VariableMappingList inputMappings = new VariableMappingList();

        [Browsable(false)]
        public VariableMappingList InputDataMappings
        {
            get
            {
                if (this.inputMappings == null)
                    this.inputMappings = new VariableMappingList();
                return inputMappings;
            }
        }

        #endregion

        #endregion

        #region validation

        private IEnumerable<ValidationBreak> BaseValidationBreaks
        {
            get { return base.GetValidationBreaks(); }
        }

        protected override IEnumerable<ValidationBreak> GetValidationBreaks()
        {
            // add base reasons
            foreach (ValidationBreak vb in this.BaseValidationBreaks)
                yield return vb;

            if (this.model != null)
            {
                // validate outbound links
                List<ISOComponentOutboundPath> unlinkedPaths = new List<ISOComponentOutboundPath>();
                ISOComponentOutboundPath[] paths = this.OutboundPaths;
                if ((paths != null) && (paths.Length > 0))
                    unlinkedPaths.AddRange(paths);

                ISOComponentLink[] links = this.model.GetOutboundLinksForComponent(this.id);
                if ((links != null) && (links.Length > 0))
                {
                    List<Guid> usedPaths = new List<Guid>();
                    foreach (ISOComponentLink link in links)
                    {
                        if (usedPaths.Contains(link.SourceComponentPath.ID) == true)
                            yield return new ValidationBreak(this,
                                "You can't have two outbound links on the same path",
                                "Link to '" + link.TargetComponent.Name +
                                "' is not the only one on '" + link.SourceComponentPath.Name +
                                "' path.");
                        else
                        {
                            usedPaths.Add(link.SourceComponentPath.ID);
                            unlinkedPaths.Remove(link.SourceComponentPath);
                        }
                    }
                }

                foreach (ISOComponentOutboundPath path in unlinkedPaths)
                    yield return new ValidationBreak(this,
                        "Path '" + path.Name + "' is not linked.",
                        "All outbound paths must be linked with another components on the same model.");

                // validate input variable mappings
                Variable[] inputVars = InputVariableAttribute.GetInputDataDefinitions(this);
                if ((inputVars != null) && (inputVars.Length > 0))
                {
                    foreach (Variable var in inputVars)
                    {
                        VariableMapping mapping = this.inputMappings.Find(
                            delegate(VariableMapping tmpMapping)
                            {
                                return (tmpMapping.Variablename == var.Name);
                            });

                        if (mapping == null)
                            this.inputMappings.Add(mapping = new VariableMapping(var.Name));

                        mapping.RunValidation();
                        if (mapping.IsValid == false)
                        {
                            foreach (ValidationBreak vb in mapping.ValidationBreaks)
                                yield return vb;
                        }
                    }
                }
            }
            else
                yield return new ValidationBreak(this,
                    "Component is not part of a model.",
                    "A ISOComponent must be part of a model in order to work.",
                    "ParentModel");
        }

        #endregion

        #region change event

        [field: NonSerialized]
        public event ComponentEvent ComponentChanged;

        protected void OnComponentChanged()
        {
            if (this.ComponentChanged != null)
                this.ComponentChanged(this);
        }

        #endregion

        public abstract ComponentOutput Run();

        public override string ToString()
        {
            return this.name;
        }
    }
}
