using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.Serialization;

using Softix.Orchestration.Core.Components;
using Softix.Orchestration.Core.Components.Events;
using Softix.Orchestration.Core.Data;
using Softix.Orchestration.Core.Models.Events;
using Softix.Orchestration.Core.Validation;

namespace Softix.Orchestration.Core.Models
{
    [ Serializable ]
    public abstract class ModelBase : ValidationBase, ISOModel, IDeserializationCallback
    {
        private List<ISOComponent> components = new List<ISOComponent>();
        private List<ISOComponentLink> links = new List<ISOComponentLink>();

        #region public String Name { get; set; }

        private String name;

        public String Name
        {
            get { return name; }
            set { name = value; }
        }

        #endregion

        #region ISOModel Members, components

        public void AddComponent(ISOComponent component)
        {
            ISOComponent tmpComp = this.GetComponent(component.ID);
            if (tmpComp == component)
                return;
            if (tmpComp != null)
                throw new InvalidOperationException("Duplicate component ID found.");

            // add component
            if (component.ParentModel != null)
                component.ParentModel.RemoveComponent(component);
            component.ParentModel = this;
            this.components.Add(component);
            this.OnComponentAdded(component);
        }

        public void RemoveComponent(Guid componentID)
        {
            ISOComponent tmpComp = this.GetComponent(componentID);
            if (tmpComp != null)
                this.RemoveComponent(tmpComp);
        }

        public void RemoveComponent(ISOComponent component)
        {
            if (component.ParentModel == this)
            {
                // delete inbound links
                ISOComponentLink[] inLinks = this.GetInboundLinksForComponent(component);
                if ((inLinks != null) && (inLinks.Length > 0))
                {
                    foreach (ISOComponentLink link in inLinks)
                        this.RemoveComponentLink(link);
                }
                // delete inbound links
                ISOComponentLink[] outLinks = this.GetOutboundLinksForComponent(component);
                if ((outLinks != null) && (outLinks.Length > 0))
                {
                    foreach (ISOComponentLink link in outLinks)
                        this.RemoveComponentLink(link);
                }
                // remove component
                component.ParentModel = null;
                this.components.Remove(component);
                this.OnComponentRemoved(component);
            }
        }

        public ISOComponent GetComponent(Guid componentID)
        {
            return this.components.Find(
                delegate(ISOComponent comp)
                {
                    return (comp.ID == componentID);
                });
        }

        #region events

        [field: NonSerialized]
        public event ModelComponentHandler ComponentAdded;
        [field: NonSerialized]
        public event ModelComponentHandler ComponentRemoved;
        [field: NonSerialized]
        public event ModelComponentHandler ComponentChanged;

        protected virtual void OnComponentAdded(ISOComponent component)
        {
            this.RunValidation();
            if (this.ComponentAdded != null)
                this.ComponentAdded(this, component);
        }

        protected virtual void OnComponentRemoved(ISOComponent component)
        {
            this.RunValidation();
            if (this.ComponentRemoved != null)
                this.ComponentRemoved(this, component);
        }

        protected virtual void OnComponentChanged(ISOComponent component)
        {
            this.RunValidation();
            if (this.ComponentChanged != null)
                this.ComponentChanged(this, component);
        }

        #endregion

        #endregion

        #region ISOModel Members, component links

        public void AddComponentLink(ISOComponentLink link)
        {
            ISOComponentLink tmpLink = this.GetComponentLink(link.ID);
            if (tmpLink == link)
                return;
            if (tmpLink != null)
                throw new InvalidOperationException("Duplicate component link found.");

            // test if source and target are in this model
            ISOComponent source = this.GetComponent(link.SourceComponent.ID);
            if (source == null)
                throw new InvalidOperationException("Source component not found in model.");
            if (Array.Find<ISOComponentOutboundPath>(source.OutboundPaths,
                delegate(ISOComponentOutboundPath tmpPath)
                {
                    return (tmpPath.ID == link.SourceComponentPath.ID);
                }) == null)
                throw new InvalidOperationException("Source component path not found.");
            if (this.GetComponent(link.TargetComponent.ID) == null)
                throw new InvalidOperationException("Target component not found.");

            // search for duplicate link
            tmpLink = this.links.Find(
                delegate(ISOComponentLink tmp)
                {
                    return (
                        (tmp.SourceComponent.ID == link.SourceComponent.ID) &&
                        (tmp.SourceComponentPath.ID == link.SourceComponentPath.ID) &&
                        (tmp.TargetComponent.ID == link.TargetComponent.ID));
                });
            if (tmpLink != null)
                throw new InvalidOperationException("Duplicate link found.");

            // add link
            if (link.ParentModel != null)
                link.ParentModel.RemoveComponentLink(link);
            link.ParentModel = this;
            this.links.Add(link);
            this.OnComponentLinkAdded(link);
            
            // validate source & target
            if (link.SourceComponent is ISOSupportsValidation)
                ((ISOSupportsValidation)link.SourceComponent).RunValidation();
            if (link.TargetComponent is ISOSupportsValidation)
                ((ISOSupportsValidation)link.TargetComponent).RunValidation();
        }

        public void RemoveComponentLink(Guid linkID)
        {
            ISOComponentLink tmpLink = this.GetComponentLink(linkID);
            if (tmpLink != null)
                this.RemoveComponentLink(tmpLink);
        }

        public void RemoveComponentLink(ISOComponentLink link)
        {
            if (link.ParentModel == this)
                link.ParentModel = null;
            this.links.Remove(link);
            this.OnComponentLinkRemoved(link);

            // validate source & target
            if (link.SourceComponent is ISOSupportsValidation)
                ((ISOSupportsValidation)link.SourceComponent).RunValidation();
            if (link.TargetComponent is ISOSupportsValidation)
                ((ISOSupportsValidation)link.TargetComponent).RunValidation();
        }

        public ISOComponentLink GetComponentLink(Guid linkID)
        {
            return this.links.Find(
                delegate(ISOComponentLink link)
                {
                    return (link.ID == linkID);
                });
        }

        #region component inbout/outbound links

        public ISOComponentLink[] GetOutboundLinksForComponent(Guid componentID)
        {
            return this.links.FindAll(
                delegate(ISOComponentLink link)
                {
                    return (link.SourceComponent.ID == componentID);
                }).ToArray();
        }

        public ISOComponentLink[] GetOutboundLinksForComponent(ISOComponent component)
        {
            return this.GetOutboundLinksForComponent(component.ID);
        }

        public ISOComponentLink[] GetOutboundLinksForComponent(Guid componentID, Guid pathID)
        {
            return this.links.FindAll(
                delegate(ISOComponentLink link)
                {
                    return ((link.SourceComponent.ID == componentID) &&
                        (link.SourceComponentPath.ID == pathID));
                }).ToArray();
        }

        public ISOComponentLink[] GetOutboundLinksForComponent(ISOComponent component, ISOComponentOutboundPath path)
        {
            return this.GetOutboundLinksForComponent(component.ID, path.ID);
        }

        public ISOComponentLink[] GetInboundLinksForComponent(Guid componentID)
        {
            return this.links.FindAll(
                delegate(ISOComponentLink link)
                {
                    return (link.TargetComponent.ID == componentID);
                }).ToArray();
        }

        public ISOComponentLink[] GetInboundLinksForComponent(ISOComponent component)
        {
            return this.GetInboundLinksForComponent(component.ID);
        }

        #endregion

        #region events

        [field: NonSerialized]
        public event ModelComponentLinkHandler ComponentLinkAdded;
        [field: NonSerialized]
        public event ModelComponentLinkHandler ComponentLinkRemoved;
        [field: NonSerialized]
        public event ModelComponentLinkHandler ComponentLinkChanged;

        protected virtual void OnComponentLinkAdded(ISOComponentLink link)
        {
            this.RunValidation();
            if (this.ComponentLinkAdded != null)
                this.ComponentLinkAdded(this, link);
        }

        protected virtual void OnComponentLinkRemoved(ISOComponentLink link)
        {
            this.RunValidation();
            if (this.ComponentLinkRemoved != null)
                this.ComponentLinkRemoved(this, link);
        }

        protected virtual void OnComponentLinkChanged(ISOComponentLink link)
        {
            this.RunValidation();
            if (this.ComponentLinkChanged != null)
                this.ComponentLinkChanged(this, link);
        }

        #endregion

        #endregion

        #region ISOModel Members, data for links

        public Variable[] GetVariablesForComponentLink(ISOComponentLink link)
        {
            return this.GetVariablesForComponentLink(link.ID);
        }

        protected void SearchVariables(Guid linkID, List<Guid> visited, VariableList variables)
        {
            ISOComponentLink link = this.GetComponentLink(linkID);
            if (link == null)
                throw new InvalidOperationException("Component link not found in model.");
            if (visited.Contains(link.ID) == true)
                return;
            visited.Add(link.ID);

            Variable[] sourceCompDefs = OutputVariableAttribute.GetOutputDataDefinitions(
                link.SourceComponent,
                link.SourceComponentPath);
                // link.SourceComponent.GetDataDefinitionsForPath(link.SourceComponentPath);
            if ((sourceCompDefs != null) && (sourceCompDefs.Length > 0))
                variables.AddRange(sourceCompDefs);

            foreach (ISOComponentLink tmpLink in this.GetInboundLinksForComponent(link.SourceComponent))
            {
                this.SearchVariables(tmpLink.ID, visited, variables);
            }
        }

        public Variable[] GetVariablesForComponentLink(Guid linkID)
        {
            VariableList list = new VariableList();
            this.SearchVariables(linkID, new List<Guid>(), list);
            return list.UniqueVariables.ToArray();
        }

        #endregion

        #region ISOModel Members, data for components

        public Variable[] GetVariablesForComponent(Guid componentID)
        {
            ISOComponent component = this.GetComponent(componentID);
            if (component == null)
                throw new InvalidOperationException("Component not found in model.");

            VariableList list = new VariableList();

            ISOComponentLink[] inbound = this.GetInboundLinksForComponent(component.ID);
            if (inbound != null)
            {
                foreach (ISOComponentLink link in inbound)
                {
                    Variable[] linkVars = this.GetVariablesForComponentLink(link);
                    if ((linkVars != null) && (linkVars.Length > 0))
                        list.AddRange(linkVars);
                }
            }

            return list.UniqueVariables.ToArray();
        }

        public Variable[] GetVariablesForComponent(ISOComponent component)
        {
            return this.GetVariablesForComponent(component.ID);
        }

        #endregion

        #region ISOModel Members, data

        #region public VariableValue[] InitialVariableValues { get; set; }

        private VariableValue[] initialVariablesValues;

        public VariableValue[] InitialVariableValues
        {
            get { return initialVariablesValues; }
            set { initialVariablesValues = value; }
        }

        #endregion

        #region public Variable[] InitialVariables { get; set; }

        private Variable[] initialVariables;

        public Variable[] InitialVariables
        {
            get { return initialVariables; }
            set
            {
                initialVariables = value;
                this.OnInitialVariablesChanged();
            }
        }

        #endregion
        #region public Variable[] FinalVariables { get; set; }

        private Variable[] finalVariables;

        public Variable[] FinalVariables
        {
            get { return finalVariables; }
            set
            {
                finalVariables = value;
                this.OnFinalVariablesChanged();
            }
        }

        #endregion
    
        [ field: NonSerialized ]
        public event ModelDataChanged InitialVariablesChanged;
        [field: NonSerialized]
        public event ModelDataChanged FinalVariablesChanged;

        protected void OnInitialVariablesChanged()
        {
            if (this.InitialVariablesChanged != null)
                this.InitialVariablesChanged(this);
        }

        protected void OnFinalVariablesChanged()
        {
            if (this.FinalVariablesChanged != null)
                this.FinalVariablesChanged(this);
        }

        #endregion

        #region validation

        private IEnumerable<ValidationBreak> BaseValidationBreaks
        {
            get { return base.GetValidationBreaks(); }
        }

        protected override IEnumerable<ValidationBreak> GetValidationBreaks()
        {
            // add base reasons
            foreach (ValidationBreak validationBreak in this.BaseValidationBreaks)
                yield return validationBreak;

            // validate initial data definitions
            if (this.initialVariables != null)
            {
                foreach (Variable var in this.initialVariables)
                {
                    // validate definitions
                    var.RunValidation();
                    if (var.IsValid == false)
                    {
                        foreach (ValidationBreak vb in var.ValidationBreaks)
                            yield return vb;
                    }
                    else
                    {
                        // validate input data
                        if ((this.initialVariablesValues == null) ||
                            (Array.Exists<VariableValue>(
                             this.initialVariablesValues,
                             delegate(VariableValue vv)
                             {
                                 return (vv.Name == var.Name);
                             }) == false))
                        {
                            yield return new ValidationBreak(var,
                                "Variable doesn't have a initial value.",
                                "You must supply values for all initial variables.");
                        }
                    }
                }
            }

            // validate final data definitions
            if (this.finalVariables != null)
            {
                foreach (Variable def in this.finalVariables)
                {
                    // validate definitions
                    def.RunValidation();
                    if (def.IsValid == false)
                    {
                        foreach (ValidationBreak vb in def.ValidationBreaks)
                            yield return vb;
                    }
                }
            }

            // validate components
            foreach (ISOComponent component in this.components)
            {
                if (component is ISOSupportsValidation)
                {
                    ISOSupportsValidation valid = (ISOSupportsValidation)component;
                    valid.RunValidation();
                    if (valid.IsValid == false)
                    {
                        foreach (ValidationBreak vb in valid.ValidationBreaks)
                            yield return vb;
                    }
                }
            }

            // validate links (??)
            foreach (ISOComponentLink link in this.links)
            {
                if (link is ISOSupportsValidation)
                {
                    ISOSupportsValidation valid = (ISOSupportsValidation)link;
                    valid.RunValidation();
                    if (valid.IsValid == false)
                    {
                        foreach (ValidationBreak vb in valid.ValidationBreaks)
                            yield return vb;
                    }
                }
            }
        }

        #endregion

        #region IDeserializationCallback Members

        public virtual void OnDeserialization(object sender)
        {
            // gigi was here
            foreach(ISOComponent component in this.components)
                component.ComponentChanged += new ComponentEvent(OnComponentChanged);
            this.RunValidation();
        }

        #endregion

        public ISOComponent[] StartComponents
        {
            get
            {
                return this.components.FindAll(
                    delegate(ISOComponent comp)
                    {
                        Variable[] inVars = InputVariableAttribute.GetInputDataDefinitions(comp);
                        if ((inVars != null) && (inVars.Length > 0))
                            return false;

                        ISOComponentLink[] inLinks = this.GetInboundLinksForComponent(comp.ID);
                        if ((inLinks != null) && (inLinks.Length > 0))
                            return false;

                        return true;
                    }).ToArray();
            }
        }

        // the one and only MUST IMPLEMENT method
        public abstract DataStorage Run(ISOComponent start);
    }
}
