using System;
using System.Drawing;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Xml.XPath;
using System.Xml.Serialization;

using Softix.Orchestration.Core;
using Softix.Orchestration.Core.Components;
using Softix.Orchestration.Core.Models;
using Softix.Orchestration.Core.Validation;
using Softix.Drawing.Diagram.Core;
using Softix.Drawing.Diagram.Simple;

namespace Softix.Orchestration.Editor.Diagram
{
    public sealed class OrchestrationDiagram : SimpleDiagram, IXmlSerializable
    {
        public OrchestrationDiagram()
        { }

        public OrchestrationDiagram(ISOModel model)
            : base()
        {
            this.model = model;
            this.model.ComponentAdded += new Softix.Orchestration.Core.Models.Events.ModelComponentHandler(ModelComponentAdded);
            this.model.ComponentRemoved += new Softix.Orchestration.Core.Models.Events.ModelComponentHandler(ModelComponentRemoved);
            this.model.ComponentLinkAdded += new Softix.Orchestration.Core.Models.Events.ModelComponentLinkHandler(ModelComponentLinkAdded);
            this.model.ComponentLinkRemoved += new Softix.Orchestration.Core.Models.Events.ModelComponentLinkHandler(ModelComponentLinkRemoved);
        }

        #region public ISOModel Model { get; }

        private ISOModel model = null;

        public ISOModel Model
        {
            get { return this.model; }
        }

        #endregion

        #region model events

        void ModelComponentAdded(ISOModel model, ISOComponent component)
        {
            IDiagramObject existingObj = Array.Find<IDiagramObject>(
                base.DiagramObjects,
                delegate(IDiagramObject tmpObj)
                {
                    return ((tmpObj is OrchestrationObject) &&
                        (((OrchestrationObject)tmpObj).Component.ID == component.ID));
                });
            if (existingObj != null)
            {
                if (existingObj is OrchestrationObject)
                    ((OrchestrationObject)existingObj).Refresh();
                return; // we already have this object
            }

            IDiagramObject obj = new OrchestrationObject(component);
            obj.Location = new Point(15, 15);
            base.AddDiagramObject(obj);
        }

        void ModelComponentLinkAdded(ISOModel model, ISOComponentLink link)
        {
            IDiagramObjectLink existingLink = Array.Find<IDiagramObjectLink>(
                base.DiagramObjectLinks,
                delegate(IDiagramObjectLink tmpLink)
                {
                    return ((tmpLink.Start is OrchestrationObject) &&
                        (((OrchestrationObject)tmpLink.Start).Component.ID == link.SourceComponent.ID) &&
                        (tmpLink.StartPort is OrchestrationObjectPort) &&
                        (((OrchestrationObjectPort)tmpLink.StartPort).Path.ID == link.SourceComponentPath.ID) &&
                        (tmpLink.End is OrchestrationObject) &&
                        (((OrchestrationObject)tmpLink.End).Component.ID == link.TargetComponent.ID));
                });
            if (existingLink != null)
                return; // this link is already present in diagram

            // find source object
            IDiagramObject source = Array.Find<IDiagramObject>(
                base.DiagramObjects,
                delegate(IDiagramObject tmpObj)
                {
                    return ((tmpObj is OrchestrationObject) &&
                        (((OrchestrationObject)tmpObj).Component.ID == link.SourceComponent.ID));
                });
            if (source == null)
                throw new InvalidOperationException("Source component not found on diagram.");

            // find source object port
            IDiagramObjectPort sourcePort = Array.Find<IDiagramObjectPort>(
                source.Ports,
                delegate(IDiagramObjectPort tmpPort)
                {
                    return ((tmpPort is OrchestrationObjectPort) &&
                        (((OrchestrationObjectPort)tmpPort).Path.ID == link.SourceComponentPath.ID));
                });
            if (sourcePort == null)
                throw new InvalidOperationException("Source port not found on diagram.");

            // find target object
            IDiagramObject target = Array.Find<IDiagramObject>(
                base.DiagramObjects,
                delegate(IDiagramObject tmpObj)
                {
                    return ((tmpObj is OrchestrationObject) &&
                        (((OrchestrationObject)tmpObj).Component.ID == link.TargetComponent.ID));
                });
            if (target == null)
                throw new InvalidOperationException("Target component not found on diagram.");

            // create & add link
            SimpleDiagramLink dLink = new SimpleDiagramLink(
                source, sourcePort, target);
            base.AddDiagramObjectLink(dLink);
        }

        void ModelComponentLinkRemoved(ISOModel model, ISOComponentLink link)
        {
            IDiagramObjectLink existingLink = Array.Find<IDiagramObjectLink>(
                base.DiagramObjectLinks,
                delegate(IDiagramObjectLink tmpLink)
                {
                    return ((tmpLink.Start is OrchestrationObject) &&
                        (((OrchestrationObject)tmpLink.Start).Component.ID == link.SourceComponent.ID) &&
                        (tmpLink.StartPort is OrchestrationObjectPort) &&
                        (((OrchestrationObjectPort)tmpLink.StartPort).Path.ID == link.SourceComponentPath.ID) &&
                        (tmpLink.End is OrchestrationObject) &&
                        (((OrchestrationObject)tmpLink.End).Component.ID == link.TargetComponent.ID));
                });
            if (existingLink != null)
            {
                base.RemoveDiagramObjectLink(existingLink);
            }
        }

        void ModelComponentRemoved(ISOModel model, ISOComponent component)
        {
            IDiagramObject existingObject = Array.Find<IDiagramObject>(
                base.DiagramObjects,
                delegate(IDiagramObject tmpObject)
                {
                    return ((tmpObject is OrchestrationObject) &&
                        (((OrchestrationObject)tmpObject).Component.ID == component.ID));
                });
            if (existingObject != null)
            {
                base.RemoveDiagramObject(existingObject);
            }
        }

        public override void SomethingDropedOnUI(object dropedItem, Point location)
        {
            if (dropedItem is ComponentInfo)
            {
                ComponentInfo info = (ComponentInfo)dropedItem;
                ISOComponent component = Activator.CreateInstance(info.Type) as ISOComponent;
                if (component != null)
                {
                    // add to diagram
                    IDiagramObject obj = new OrchestrationObject(component);
                    obj.Location = location;
                    base.AddDiagramObject(obj);

                    // add to model
                    this.model.AddComponent(component);

                    return;
                }
            }
            
            base.SomethingDropedOnUI(dropedItem, location);
        }

        public override void NewLinkFromUI(IDiagramObject source, IDiagramObjectPort sourcePort, DiagramObjectPortAlign sourcePortAlign, IDiagramObject targetObject, DiagramObjectEdge targetEdge)
        {
            if ((source is OrchestrationObject) &&
                (targetObject is OrchestrationObject) &&
                (sourcePort is OrchestrationObjectPort))
            {
                ISOComponent sourceComponent = ((OrchestrationObject)source).Component;
                Guid sourcePathID = ((OrchestrationObjectPort)sourcePort).Path.ID;
                ISOComponentOutboundPath sourcePath = Array.Find<ISOComponentOutboundPath>(
                    sourceComponent.OutboundPaths,
                    delegate(ISOComponentOutboundPath tmpPath)
                    {
                        return (tmpPath.ID == sourcePathID);
                    });
                if (sourcePath != null)
                {
                    ISOComponent targetComponent = ((OrchestrationObject)targetObject).Component;

                    // create model link
                    SimpleComponentLink mLink = new SimpleComponentLink();
                    mLink.SetSourceComponent(sourceComponent, sourcePath.ID);
                    mLink.TargetComponent = targetComponent;

                    // add diagram link
                    IDiagramObjectLink dLink = new OrchestrationObjectLink(
                        this, mLink);
                    dLink.StartPortAlign = sourcePortAlign;
                    dLink.EndEdge = targetEdge;
                    base.AddDiagramObjectLink(dLink);

                    // add model link
                    try
                    {
                        this.model.AddComponentLink(mLink);
                    }
                    catch (Exception ex)
                    {
                        base.RemoveDiagramObjectLink(dLink);
                        base.ErrorMessage = "[" + ex.GetType().Name + "]\n" + ex.Message;
                    }
                }
            }
        }

        #endregion

        #region IXmlSerializable Members

        public System.Xml.Schema.XmlSchema GetSchema()
        {
            return null;
        }

        public void ReadXml(XmlReader r)
        {
            XPathDocument doc = new XPathDocument(r);
            XPathNavigator nav = doc.CreateNavigator();
            if (nav.MoveToChild("Components", "") == true)
            {
                #region read component

                if (nav.MoveToChild("Component", "") == true)
                {
                    do
                    {
                        nav.MoveToAttribute("ComponentID", "");
                        string strCompID = nav.Value;
                        nav.MoveToParent();

                        Guid compID = new Guid(strCompID);
                        ISOComponent component = this.model.GetComponent(compID);

                        IDiagramObject obj = null;
                        if (component != null)
                        {
                            obj = new OrchestrationObject(component);
                        }
                        else
                        {
                            obj = new SimpleDiagramObject();
                            nav.MoveToChild("ComponentName", "");
                            obj.Text = "[" + nav.Value + "]";
                            nav.MoveToParent();
                        }

                        if (obj != null)
                        {
                            nav.MoveToChild("Location", "");
                            {
                                nav.MoveToAttribute("X", "");
                                int x = Int32.Parse(nav.Value);
                                nav.MoveToParent();
                                nav.MoveToAttribute("Y", "");
                                int y = Int32.Parse(nav.Value);
                                nav.MoveToParent();
                                obj.Location = new Point(x, y);
                            }
                            nav.MoveToParent();

                            base.AddDiagramObject(obj);
                        }
                    } while (nav.MoveToNext("Component", "") == true);
                    nav.MoveToParent();
                }

                #endregion
                nav.MoveToParent();
            }

            if (nav.MoveToChild("Links", "") == true)
            {
                #region read links

                if (nav.MoveToChild("Link", "") == true)
                {
                    do
                    {
                        nav.MoveToAttribute("LinkID", "");
                        string strLinkID = nav.Value;
                        nav.MoveToParent();

                        Guid linkID = new Guid(strLinkID);
                        ISOComponentLink link = this.model.GetComponentLink(linkID);

                        if (link != null)
                        {
                            IDiagramObjectLink lnk = new OrchestrationObjectLink(this, link);
                            nav.MoveToChild("LinkSource", "");
                            {
                                nav.MoveToChild("SourcePortLocation", "");
                                {
                                    DiagramObjectPortAlign portAlign = (DiagramObjectPortAlign)
                                        Enum.Parse(typeof(DiagramObjectPortAlign), nav.Value);
                                    lnk.StartPortAlign = portAlign;
                                }
                                nav.MoveToParent();
                            }
                            nav.MoveToParent();

                            nav.MoveToChild("LinkTarget", "");
                            {
                                nav.MoveToChild("TargetEdge", "");
                                {
                                    DiagramObjectEdge edge = (DiagramObjectEdge)
                                        Enum.Parse(typeof(DiagramObjectEdge), nav.Value);
                                    lnk.EndEdge = edge;
                                }
                                nav.MoveToParent();
                            }
                            nav.MoveToParent();

                            base.AddDiagramObjectLink(lnk);
                        }
                    } while (nav.MoveToNext("Link", "") == true);
                    nav.MoveToParent();
                }

                #endregion
                nav.MoveToParent();
            }

            if (this.model is ISOSupportsValidation)
                ((ISOSupportsValidation)this.model).RunValidation();
        }

        public void WriteXml(XmlWriter w)
        {
            w.WriteAttributeString("ModelName", this.model.Name);

            w.WriteStartElement("Components");
            {
                foreach (OrchestrationObject obj in
                    Array.FindAll<IDiagramObject>(this.DiagramObjects,
                    delegate(IDiagramObject ido)
                    {
                        return (ido is OrchestrationObject);
                    }))
                {
                    obj.WriteXml(w);
                }
            }
            w.WriteEndElement();

            w.WriteStartElement("Links");
            {
                foreach (OrchestrationObjectLink lnk in
                    Array.FindAll<IDiagramObjectLink>(this.DiagramObjectLinks,
                    delegate(IDiagramObjectLink idol)
                    {
                        return (idol is OrchestrationObjectLink);
                    }))
                {
                    lnk.WriteXml(w);
                }
            }
            w.WriteEndElement();
        }

        #endregion
    }
}
