using System;
using System.Collections.Generic;
using System.Text;
using System.Xml.Serialization;

using Softix.Drawing.Diagram.Core;
using Softix.Drawing.Diagram.Simple;
using Softix.Orchestration.Core.Components;
using Softix.Orchestration.Core.Validation;
using Softix.Orchestration.Editor.Editors;

namespace Softix.Orchestration.Editor.Diagram
{
    public sealed class OrchestrationObject : SimpleDiagramObject, IDiagramHasActions, IXmlSerializable
    {
        public OrchestrationObject()
            : base()
        { }

        public OrchestrationObject(ISOComponent component)
            : base()
        {
            this.Component = component;
        }

        #region public ISOComponent Component { get; set; }

        private ISOComponent component;

        public ISOComponent Component
        {
            get { return component; }
            set
            {
                component = value;
                this.component.ComponentChanged += new Softix.Orchestration.Core.Components.Events.ComponentEvent(Component_ComponentChanged);
                if (this.component is ISOSupportsValidation)
                {
                    ((ISOSupportsValidation)this.component).ValidationChanged += new ValidationHandler(OrchestrationObject_ValidationChanged);
                }
                this.Refresh();
            }
        }

        void Component_ComponentChanged(ISOComponent component)
        {
            if (component == this.component)
                Refresh();
        }

        void OrchestrationObject_ValidationChanged(ISOSupportsValidation source)
        {
            if (source == this.component)
                Refresh();
        }

        #endregion

        internal void Refresh()
        {
            if (this.component == null)
            {
                base.ID = Guid.Empty;
                base.ClearPorts();
                base.Text = "(no component)";
                base.Name = "";
            }
            else
            {
                // SMART update of ports
                Dictionary<Guid, IDiagramObjectPort> portsToRemove =
                    new Dictionary<Guid, IDiagramObjectPort>();
                foreach (IDiagramObjectPort port in base.Ports)
                {
                    if (port is OrchestrationObjectPort)
                        portsToRemove.Add(((OrchestrationObjectPort)port).Path.ID, port);
                    else
                        portsToRemove.Add(Guid.NewGuid(), port);
                }
                foreach (ISOComponentOutboundPath path in this.component.OutboundPaths)
                {
                    if (portsToRemove.ContainsKey(path.ID) == true)
                    {
                        // existing port
                        if (portsToRemove[path.ID] is OrchestrationObjectPort)
                        {
                            OrchestrationObjectPort ooPort = (OrchestrationObjectPort)portsToRemove[path.ID];
                            ooPort.Refresh();
                            portsToRemove.Remove(path.ID);
                        }
                    }
                    else
                    {
                        // new port
                        base.AddPort(new OrchestrationObjectPort(path));
                    }
                }
                // remove ports
                foreach (KeyValuePair<Guid, IDiagramObjectPort> pair in portsToRemove)
                {
                    base.ports.Remove(pair.Value);
                }

                // update name and text
                base.Name = this.component.Name;
                base.Text = this.component.Name;

                // update artifacts
                base.ClearArtifacts();
                if (this.component is ISOSupportsValidation)
                {
                    ISOSupportsValidation valid = (ISOSupportsValidation)this.component;
                    if (valid.IsValid == false)
                        base.AddArtifact("exception", SimpleDiagramArtifact.ExceptionArtifact);
                }
                if (this.component.ParentModel != null)
                {
                    ISOComponentLink[] inbound = this.component.ParentModel.GetInboundLinksForComponent(this.component.ID);
                    if ((inbound == null) || (inbound.Length == 0))
                        base.AddArtifact("start", SimpleDiagramArtifact.StartArtifact);
                    
                    ISOComponentLink[] outbound = this.component.ParentModel.GetOutboundLinksForComponent(this.component.ID);
                    ISOComponentOutboundPath[] paths = this.component.OutboundPaths;
                    if (
                        ((outbound == null) || (outbound.Length == 0)) &&
                        ((paths == null) || (paths.Length == 0))
                        )
                        base.AddArtifact("end", SimpleDiagramArtifact.EndArtifact);
                }
            }
        }

        #region IDiagramHasActions Members

        public Dictionary<String, IDiagramAction> GetActions()
        {
            Dictionary<String, IDiagramAction> dic = new Dictionary<String, IDiagramAction>();

            dic.Add("data", new SimpleDiagramAction("Browse Data"));
            dic.Add("edit", new SimpleDiagramAction("Edit Component"));

            bool gigi = false;
            if (this.component is ISOSupportsValidation)
            {
                ISOSupportsValidation valid = (ISOSupportsValidation)this.component;
                if (valid.IsValid == false)
                    gigi = true;
            }
            dic.Add("breaks", new SimpleDiagramAction("View Invalid Reasons", gigi));

            dic.Add("SEP", new SimpleDiagramAction("-"));
            dic.Add("delete", new SimpleDiagramAction("Delete ME!"));

            return dic;
        }

        public void PerformAction(string key)
        {
            switch (key)
            {
                case "data":
                    using (BrowseDataForm form = new BrowseDataForm(this.component))
                    {
                        form.ShowDialog();
                    }
                    break;
                case "edit":
                    using (ComponentEditor editor = new ComponentEditor(this.component))
                    {
                        editor.ShowDialog();
                    }
                    break;
                case "breaks":
                    using (ValidationBreaksViewer viewer = new ValidationBreaksViewer(
                        ((ISOSupportsValidation)this.component).ValidationBreaks))
                    {
                        viewer.ShowDialog();
                    }
                    break;
                case "delete":
                    if (this.component.ParentModel != null)
                        this.component.ParentModel.RemoveComponent(this.component.ID);
                    break;
            }
        }

        #endregion

        #region IXmlSerializable Members

        public System.Xml.Schema.XmlSchema GetSchema()
        {
            return null;
        }

        public void ReadXml(System.Xml.XmlReader r)
        {

        }

        public void WriteXml(System.Xml.XmlWriter w)
        {
            w.WriteStartElement("Component");
            w.WriteAttributeString("ComponentID", this.component.ID.ToString());
            w.WriteElementString("ComponentName", this.component.Name);

            w.WriteStartElement("Location");
            w.WriteAttributeString("X", base.Location.X.ToString());
            w.WriteAttributeString("Y", base.Location.Y.ToString());
            w.WriteEndElement();

            w.WriteEndElement();
        }

        #endregion
    }
}
